# -*- coding: utf-8 -*-

import numpy as np
from itertools import product

from .facility_location_solution import FacilityLocationSolution
from ..stochastic_problem_basis import StochasticProblemBasis

class FacilityLocationProblemMultistage(StochasticProblemBasis):
    """Three-Stage Facility Location Problem.
    
    Argument:
    ---------
    param: dict with keys and values:
        "pos_client":               2d-array of shape (n_client_locations, 2)
        "pos_facility":             2d-array of shape (n_facility_locations, 2)
        "opening_cost":             1d-array of shape (n_facility_locations,)
        "facility_capacity":        1d-array of shape (n_facility_locations,)
        "penalty":                  1d-array of shape (n_facility_locations,)
        "shipping_cost":            (optional) 2d-array of shape (n_client_locations, n_facility_locations)
        "resources":                (optional) 2d-array of shape (n_client_locations, n_facility_locations)
        "max_facilities":           int
        "min_facilities_in_zone":   1d-array of shape (n_zones,)
        "facility_in_zone":         1d-array of shape (n_facility_locations,) (of int values in [0, n_zones-1])
        
    If the keys "shipping_cost" and/or "resources" are not provided, they are computed 
    from the distances between facilities and clients (see methods: resources() and 
    shipping_cost()).
    """
        
    def __init__(self, param):                
        self.param = param
        self.n_facility_locations = self.param['pos_facility'].shape[0]
        self.n_client_locations = self.param['pos_client'].shape[0]
        self.n_zones = self.param['min_facilities_in_zone'].shape[0]
                
        # sets
        self.client_locations = range(self.n_client_locations)
        self.facility_locations = range(self.n_facility_locations)
        self.zones = range(self.n_zones)
        
        # compute the distances between clients and facilities
        # this distance is the resource consumed by client c served by facility f
        self.dist = 30 * np.linalg.norm(self.pos_client()[:, np.newaxis, :] \
                                     - self.pos_facility()[np.newaxis, :, :], axis=2)
        
        # stochastic problem information
        StochasticProblemBasis.__init__(self, 
                                        name='Facility Location Problem Multistage',
                                        n_stages=3,
                                        objective_sense='min',
                                        is_obj_random=False,
                                        is_mip=True,
                                        solution_class=FacilityLocationSolution)

    # --- Problem definition ---
    def decision_variables_definition(self, stage):
        """For each decision variables of type {y_{t,i} \in A : i \in I} where I is an indexing set 
        (a list of tuples), y_{t,i} is of type A (with A = binary (B), or integer (I), or continuous (C)), 
        and lb <= y_{t,i} <= ub, this function generates a 5-tuple of the form: ('y', I, lb, ub, A) 
        under the conditon: stage == t. (lb and ub can be two lists provided they have the same size as I)"""
        if stage == 0:
            yield 'x', self.facility_locations, 0, 1, 'B'
        elif stage == 1:
            yield 'y', list(product(self.client_locations, self.facility_locations)), 0, 1, 'B'
        elif stage == 2:
            yield 'z', self.facility_locations, 0, None, 'C'
            
    def random_variables_definition(self, stage):
        """For each random variable of type {xi_{stage,i} : i \in I} where I is an indexing set (an iterable), 
        this method generates a 2-tuple of the form ('xi', I)"""
        if stage == 1:
            yield 'h', self.client_locations
        elif stage == 2:
            yield 'd', self.client_locations
    
    def objective(self):
        """Returns the problem's objective function"""
        return self.dot(self.x(), self.opening_cost()) \
                + self.dot(self.y(), self.shipping_cost().flatten()) \
                + self.dot(self.z(), self.penalty())
    
    def deterministic_linear_constraints(self, stage):
        """Generates all the problem's constraints that do not depend on the random parameters"""
        if stage == 0:
            yield self.max_facility_constraint()
            yield self.min_facility_in_zone_constraint()
        if stage == 1:
            pass
        elif stage == 2:
            yield self.resource_consumption_constraint()

    def random_linear_constraints(self, stage):
        """Generates all the problem's constraints that depend on the random parameters"""
        if stage == 1: 
            yield self.assignment_client_constraint()

    # ---- Decision variables ----    
    def x(self, f=None):
        """x_{f} = 1 if facility f is open, 0 otherwise"""
        return self.get_dvar(0, 'x', f)
    
    def y(self, c=None, f=None):
        """y_{c,f} = 1 if facility f ships to client c, 0 otherwise"""
        if c is None and f is None: 
            return self.get_dvar(1, 'y')
        elif c is None and f is not None: 
            return self.get_dvar(1, 'y').reshape(-1, self.n_facility_locations)[:, f]
        elif c is not None and f is None: 
            return self.get_dvar(1, 'y').reshape(self.n_client_locations, -1)[c, :]
        else: 
            return self.get_dvar(1, 'y', (c, f))
         
    def z(self, f=None):
        """z_{f} = penalty for overloading facility f"""
        return self.get_dvar(2, 'z', f)
    
    # ---- Random variables ----       
    def h(self, c=None):
        """h_{c} = 1 if client c is present in scenario s, 0 otherwise"""
        return self.get_rvar(1, 'h', c)
    
    def d(self, c=None):
        """d_{c}: demand form client c"""
        return self.get_rvar(2, 'd', c)
    
    # --- Precomputation ---
    def precompute_decision_variables(self, stage):
        pass
    
    def precompute_parameters(self, stage):
        pass
    
    # --- Sanity check ---
    def sanity_check(self, stage):
        pass
    
    # --- Parameters ---
    def pos_client(self, c=None):
        """Position of client c"""
        pos = np.array(self.param["pos_client"])
        return pos if c is None else pos[c]
        
    def pos_facility(self, f=None):
        """Position of facility f"""
        pos = np.array(self.param["pos_facility"])
        return pos if f is None else pos[f]
        
    def opening_cost(self, f=None):
        """Cost of locating a facility f"""
        cost = np.array(self.param["opening_cost"])
        return cost if f is None else cost[f]
    
    def capacity(self, f=None):
        """Facility capacity"""
        return self.param['facility_capacity'][f]
    
    def shipping_cost(self, c=None, f=None):
        if self.param.get('shipping_cost') is None:
            return self.distance(c, f) - 1
        else:
            if c is None and f is None: return self.param.get('shipping_cost')
            elif c is None and f is not None: return self.param.get('shipping_cost')[:, f]
            elif c is not None and f is None: return self.param.get('shipping_cost')[c, :]
            else: return self.param.get('shipping_cost')[c, f]
    
    def resources(self, c=None, f=None):
        """Resources consumed by facility f shipping to client c"""
        if self.param.get('resources') is None:
            return self._resources_from_distance(c, f)
        else:
            return self._resource_from_input_array(c, f)
    
    def _resources_from_distance(self, c=None, f=None):
        """Resources consumed by facility f shipping to client c"""
        if c is None and f is None:
            return self.distance(c, f) + self.d(c)[:, np.newaxis]
        else:
            return self.distance(c, f) + self.d(c)
            
    def _resource_from_input_array(self, c=None, f=None):
        if c is None and f is None: return self.param.get('resources') + self.d(c)[:, np.newaxis]
        elif c is None and f is not None: return self.param.get('resources')[:, f] + self.d(c)
        elif c is not None and f is None: return self.param.get('resources')[c, :] + self.d(c)
        else: return self.param.get('resources')[c, f] + self.d(c)
    
    def max_facilities(self):
        """Upper bound on the total number of facilities that can be located"""
        return self.param["max_facilities"]
    
    def min_facilities(self, z):
        """Minimum number of facilities to be located in zone z"""
        return self.param["min_facilities_in_zone"][z]
    
    def penalty(self, f=None):
        """Penality for exceeding facility capacity"""
        penalty = np.array(self.param["penalty"])
        return penalty if f is None else penalty[f]
    
    def is_in_zone(self, z):
        """Return a 1d-array with True at pos k if facility k is in zone z"""
        return np.array(self.param["facility_in_zone"]) == z
            
    def distance(self, c=None, f=None):
        """Distance between client location c and facility location f"""
        if c is None and f is None: return self.dist
        elif c is None and f is not None: return self.dist[:, f]
        elif c is not None and f is None: return self.dist[c, :]
        else: return self.dist[c, f]
        
    # --- constraints ---
    def max_facility_constraint(self):
        yield self.sum(self.x()) <= self.max_facilities(), "max_facilities"
        
    def min_facility_in_zone_constraint(self):
        for zone in self.zones:
            yield self.sum(self.x()[self.is_in_zone(zone)]) >= self.min_facilities(zone), f"min_facilities_zone_{zone}"
             
    def resource_consumption_constraint(self):
        self.big_M = 10**10
        for f in self.facility_locations:
            yield self.dot(self.y(f=f), self.resources(f=f)) - self.z(f) <= self.capacity(f) * self.x(f), \
                    f"resource_facility_{f}"
            for c in self.client_locations:
                yield self.y(c, f) <= self.x(f)
            yield self.z(f) <= self.big_M * self.x(f), f"big_M_penalty_facility_{f}_1"
            
    def assignment_client_constraint(self):
        for c in self.client_locations:
            yield self.sum(self.y(c=c)) == self.h(c), f"assignment_client_{c}"
            
    # --- Load, save ---
    @classmethod
    def from_file(cls, path, extension='txt'):
        if extension == 'txt':
            with open(f'{path}.txt', "r") as f:
                file_str = f.read()
                file_str = file_str.replace('array', 'np.array')
                file_str = file_str.replace('nan', 'np.nan')
                param = eval(file_str)
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.pickle', "rb") as f:
                param = pickle.load(f)
        return cls(param)
    
    def to_file(self, path, extension='txt'):
        if extension == 'txt':
            np.set_printoptions(threshold=np.inf) # no limit on the number of elements printed in an array
            with open(f'{path}.txt', "w") as f:
                f.write(repr(self.param).replace("),", "),\n"))
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.pickle', "wb") as f:
                pickle.dump(self.param, f)
        else:
            TypeError(f"Extension should be 'pickle' or 'txt', not {extension}.")

    # --- Representation ---
    def __repr__(self):
        string_problem = StochasticProblemBasis.__repr__(self)
        string = ("Network: \n"
                  f"  {self.n_facility_locations} facility locations\n"
                  f"  {self.n_client_locations} client locations\n"
                  f"  {self.n_zones} zones")
        return string_problem + "\n" + string
    
    
class FacilityLocationProblemMultistageCapacity(StochasticProblemBasis):
    """Three-Stage Facility Location Problem with capacity decisions at 2nd stage.
    
    Argument:
    ---------
    param: dict with keys and values:
        "pos_client":               2d-array of shape (n_client_locations, 2)
        "pos_facility":             2d-array of shape (n_facility_locations, 2)
        "opening_cost":             1d-array of shape (n_facility_locations,)
        "penalty":                  1d-array of shape (n_facility_locations,)
        "shipping_cost":            (optional) 2d-array of shape (n_client_locations, n_facility_locations)
        "resources":                (optional) 2d-array of shape (n_client_locations, n_facility_locations)
        "max_facilities":           int
        "min_facilities_in_zone":   1d-array of shape (n_zones,)
        "facility_in_zone":         1d-array of shape (n_facility_locations,) (of int values in [0, n_zones-1])
        
    If the keys "shipping_cost" and/or "resources" are not provided, they are computed 
    from the distances between facilities and clients (see methods: resources() and 
    shipping_cost()).
    """
        
    def __init__(self, param):
                
        self.param = param
        self.n_facility_locations = self.param['pos_facility'].shape[0]
        self.n_client_locations = self.param['pos_client'].shape[0]
        self.n_zones = self.param['min_facilities_in_zone'].shape[0]
                
        # sets
        self.client_locations = range(self.n_client_locations)
        self.facility_locations = range(self.n_facility_locations)
        self.zones = range(self.n_zones)
        
        # compute the distances between clients and facilities
        # this distance is the resource consumed by client c served by facility f
        self.dist = 30 * np.linalg.norm(self.pos_client()[:, np.newaxis, :] \
                                     - self.pos_facility()[np.newaxis, :, :], axis=2)
        
        # stochastic problem information
        StochasticProblemBasis.__init__(self, 
                                        name='Facility Location Problem Multistage',
                                        n_stages=3,
                                        objective_sense='min',
                                        is_obj_random=False,
                                        is_mip=True,
                                        solution_class=FacilityLocationSolution)

    # --- Problem definition ---
    def decision_variables_definition(self, stage):
        """For each decision variables of type {y_{t,i} \in A : i \in I} where I is an indexing set 
        (a list of tuples), y_{t,i} is of type A (with A = binary (B), or integer (I), or continuous (C)), 
        and lb <= y_{t,i} <= ub, this function generates a 5-tuple of the form: ('y', I, lb, ub, A) 
        under the conditon: stage == t. (lb and ub can be two lists provided they have the same size as I)"""
        if stage == 0:
            yield 'x', self.facility_locations, 0, 1, 'B'
        elif stage == 1:
            yield 'u', self.facility_locations, 0, None, 'C'
        elif stage == 2:
            yield 'y', list(product(self.client_locations, self.facility_locations)), 0, 1, 'B'
            yield 'z', self.facility_locations, 0, None, 'C'
            yield 'z2', self.facility_locations, 0, None, 'C'
            
    def random_variables_definition(self, stage):
        """For each random variable of type {xi_{stage,i} : i \in I} where I is an indexing set (an iterable), 
        this method generates a 2-tuple of the form ('xi', I)"""
        if stage == 1:
            yield 'h', self.client_locations
        elif stage == 2:
            yield 'd', self.client_locations
    
    def objective(self):
        """Returns the problem's objective function"""
        return self.dot(self.x(), self.opening_cost()) \
                + self.dot(self.y(), self.shipping_cost().flatten()) \
                + self.dot(self.z() + self.z2(), self.penalty())
    
    def deterministic_linear_constraints(self, stage):
        """Generates all the problem's constraints that do not depend on the random parameters"""
        if stage == 0:
            yield self.max_facility_constraint()
            yield self.min_facility_in_zone_constraint()
        if stage == 1:
            pass
        elif stage == 2:
            pass
        
    def random_linear_constraints(self, stage):
        """Generates all the problem's constraints that depend on the random parameters"""
        if stage == 2: 
            yield self.assignment_client_constraint()
            yield self.resource_consumption_constraint()

    # ---- Decision variables ----    
    def x(self, f=None):
        """x_{f} = 1 if facility f is open, 0 otherwise"""
        return self.get_dvar(0, 'x', f)
    
    def u(self, f=None):
        """Facility capacity"""
        return self.get_dvar(1, 'u', f)
    
    def y(self, c=None, f=None):
        """y_{c,f} = 1 if facility f ships to client c, 0 otherwise"""
        if c is None and f is None: 
            return self.get_dvar(2, 'y')
        elif c is None and f is not None: 
            return self.get_dvar(2, 'y').reshape(-1, self.n_facility_locations)[:, f]
        elif c is not None and f is None: 
            return self.get_dvar(2, 'y').reshape(self.n_client_locations, -1)[c, :]
        else: 
            return self.get_dvar(2, 'y', (c, f))
         
    def z(self, f=None):
        """z_{f} = penalty for overloading facility f"""
        return self.get_dvar(2, 'z', f)
      
    def z2(self, f=None):
        """z_{f} = penalty for underloading facility f"""
        return self.get_dvar(2, 'z2', f)
    
    # ---- Random variables ----       
    def h(self, c=None):
        """h_{c} = 1 if client c is present in scenario s, 0 otherwise"""
        return self.get_rvar(1, 'h', c)
    
    def d(self, c=None):
        """d_{c}: demand form client c"""
        return self.get_rvar(2, 'd', c)
    
    # --- Precomputation ---
    def precompute_decision_variables(self, stage):
        pass
    
    def precompute_parameters(self, stage):
        pass
    
    # --- Sanity check ---
    def sanity_check(self, stage):
        pass
    
    # --- Parameters ---
    def pos_client(self, c=None):
        """Position of client c"""
        pos = np.array(self.param["pos_client"])
        return pos if c is None else pos[c]
        
    def pos_facility(self, f=None):
        """Position of facility f"""
        pos = np.array(self.param["pos_facility"])
        return pos if f is None else pos[f]
        
    def opening_cost(self, f=None):
        """Cost of locating a facility f"""
        cost = np.array(self.param["opening_cost"])
        return cost if f is None else cost[f]
    
    def shipping_cost(self, c=None, f=None):
        if self.param.get('shipping_cost') is None:
            return self.distance(c, f) - 1
        else:
            if c is None and f is None: return self.param.get('shipping_cost')
            elif c is None and f is not None: return self.param.get('shipping_cost')[:, f]
            elif c is not None and f is None: return self.param.get('shipping_cost')[c, :]
            else: return self.param.get('shipping_cost')[c, f]
    
    def resources(self, c=None, f=None):
        """Resources consumed by facility f shipping to client c"""
        if self.param.get('resources') is None:
            return self._resources_from_distance(c, f)
        else:
            return self._resource_from_input_array(c, f)
    
    def _resources_from_distance(self, c=None, f=None):
        """Resources consumed by facility f shipping to client c"""
        if c is None and f is None:
            return self.distance(c, f) + self.d(c)[:, np.newaxis]
        else:
            return self.distance(c, f) + self.d(c)
            
    def _resource_from_input_array(self, c=None, f=None):
        if c is None and f is None: return self.param.get('resources') + self.d(c)[:, np.newaxis]
        elif c is None and f is not None: return self.param.get('resources')[:, f] + self.d(c)
        elif c is not None and f is None: return self.param.get('resources')[c, :] + self.d(c)
        else: return self.param.get('resources')[c, f] + self.d(c)
            
    def max_facilities(self):
        """Upper bound on the total number of facilities that can be located"""
        return self.param["max_facilities"]
    
    def min_facilities(self, z):
        """Minimum number of facilities to be located in zone z"""
        return self.param["min_facilities_in_zone"][z]
    
    def penalty(self, f=None):
        """Penality for exceeding facility capacity"""
        penalty = np.array(self.param["penalty"])
        return penalty if f is None else penalty[f]
    
    def is_in_zone(self, z):
        """Return a 1d-array with True at pos k if facility k is in zone z"""
        return np.array(self.param["facility_in_zone"]) == z
            
    def distance(self, c=None, f=None):
        """Distance between client location c and facility location f"""
        if c is None and f is None: return self.dist
        elif c is None and f is not None: return self.dist[:, f]
        elif c is not None and f is None: return self.dist[c, :]
        else: self.dist[c, f]
        
    # --- constraints ---
    def max_facility_constraint(self):
        yield self.sum(self.x()) <= self.max_facilities(), "max_facilities"
        
    def min_facility_in_zone_constraint(self):
        for zone in self.zones:
            yield self.sum(self.x()[self.is_in_zone(zone)]) >= self.min_facilities(zone), f"min_facilities_zone_{zone}"
             
    def resource_consumption_constraint(self):
        self.big_M = 10**10
        for f in self.facility_locations:
            yield self.dot(self.y(f=f), self.resources(f=f)) - self.z(f) + self.z2(f) == self.u(f), \
                    f"resource_facility_{f}"
            for c in self.client_locations:
                yield self.y(c, f) <= self.x(f)
            yield self.z(f) <= self.big_M * self.x(f), f"big_M_penalty_facility_{f}_1"
            yield self.z2(f) <= self.big_M * self.x(f), f"big_M_penalty_facility_{f}_2"
            yield self.u(f) <= self.big_M * self.x(f), f"big_M_capacity_facility_{f}"
            
    def assignment_client_constraint(self):
        for c in self.client_locations:
            yield self.sum(self.y(c=c)) == self.h(c), f"assignment_client_{c}"
            
    # --- Load, save ---
    @classmethod
    def from_file(cls, path, extension='txt'):
        if extension == 'txt':
            with open(f'{path}.txt', "r") as f:
                file_str = f.read()
                file_str = file_str.replace('array', 'np.array')
                file_str = file_str.replace('nan', 'np.nan')
                param = eval(file_str)
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.pickle', "rb") as f:
                param = pickle.load(f)
        return cls(param)
    
    def to_file(self, path, extension='txt'):
        if extension == 'txt':
            np.set_printoptions(threshold=np.inf) # no limit on the number of elements printed in an array
            with open(f'{path}.txt', "w") as f:
                f.write(repr(self.param).replace("),", "),\n"))
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.pickle', "wb") as f:
                pickle.dump(self.param, f)
        else:
            TypeError(f"Extension should be 'pickle' or 'txt', not {extension}.")

    # --- Representation ---
    def __repr__(self):
        string_problem = StochasticProblemBasis.__repr__(self)
        string = ("Network: \n"
                  f"  {self.n_facility_locations} facility locations\n"
                  f"  {self.n_client_locations} client locations\n"
                  f"  {self.n_zones} zones")
        return string_problem + "\n" + string
    
    
from_file = FacilityLocationProblemMultistage.from_file
from_file = FacilityLocationProblemMultistageCapacity.from_file

def generate_random_parameters(n_facility_locations, 
                               n_client_locations, 
                               n_zones, 
                               with_capacity=True,
                               seed=None):
    """Generate randomly a set of deterministic parameters of the facility location problem"""
    if seed is not None:
        np.random.seed(seed)
    return {"pos_client": np.random.uniform(0, 1, size=(n_client_locations, 2)),
            "pos_facility": np.random.uniform(0, 1, size=(n_facility_locations, 2)),
            "opening_cost": np.random.randint(40, 81, size=n_facility_locations),
            "max_facilities": n_facility_locations,
            "facility_capacity": np.random.randint(30, 60, size=n_facility_locations) if with_capacity else None,
            "min_facilities_in_zone": np.array([1] * n_zones),
            "facility_in_zone": np.random.choice(range(n_zones), size=n_facility_locations),
            "penalty": 100 * np.ones(n_facility_locations)}
        


# -*- coding: utf-8 -*-

import numpy as np

import copy
import itertools
from time import time
import matplotlib.pyplot as plt

from stochoptim.scengen.tree_structure import Node
from stochoptim.scengen.scenario_tree import ScenarioTree
Cartesian = itertools.product

class TreeSearch:
    
    def __init__(self, scenario_process, variability_process, demerit, nber_stages):
        self.scenario_process = scenario_process
        self.variability_process = variability_process
        self.demerit = demerit
        self.nber_stages = nber_stages
        self.last_stage = nber_stages - 1
        
        self._search_methods = [("VNS", "forward"), ("VNS", "backward"), 
                                ("EXH", "forward"), ("EXH", "backward")]
        self._best_fod = {}
        self._best_tree = {}
        self._fod_sample = {}
        self._initialize()
        
    def _initialize(self, method=None):
        methods = self._search_methods if method is None else [method]
        for method in methods:
            self._best_fod[method] = np.inf
            self._best_tree[method] = None
            self._fod_sample[method] = []
        
    def fod_sample(self, method):
        return np.array(self._fod_sample[method])
    
    def best_tree(self, method):
        return self._best_tree[method]
    
    def best_fod(self, method):
        return self._best_fod[method]
    
    def plot_fod_hist(self, method, bins=10, figsize=(5,5)):
        fig, ax = plt.subplots(figsize=figsize)
        ax.hist(self._fod_sample[method], bins=bins)
        ax.set_title(f"P10: {np.quantile(self._fod_sample[method], 0.1):.5f} "
                     f"; P50: {np.quantile(self._fod_sample[method], 0.5):.5f}\n"
                     f"min: {min(self._fod_sample[method]):.5f} ;"
                     f"max: {max(self._fod_sample[method]):.5f}")
        plt.show()
        
    def plot_fod_progress(self, method, figsize=(5,5)):
        fig, ax = plt.subplots(figsize=(5,5))
        ax.plot(range(len(self._fod_sample[method])), np.minimum.accumulate(self._fod_sample[method]))
        best_index = np.where(np.minimum.accumulate(self._fod_sample[method]) <= self._best_fod[method] + 10**-10)[0][0]
        ax.scatter(best_index, self._best_fod[method])
        plt.show()
        
    def variable_neighborhood_search(self, 
                                    nber_scenarios, 
                                    initial_tree=None,
                                    optimized='forward', 
                                    max_iteration=np.inf,
                                    max_no_improvement=np.inf,
                                    num_local_samples=10, 
                                    num_neighborhoods=10, 
                                    neighborhood_shrink=0):
        """Explore the space of tree structures via a strategy of 'variable neighborhood' to find the scenario tree 
        of lowest demerit"""
        assert optimized in ["forward", "backward"], ("`optimized` must be either 'forward' or "
        f"'backward', not {optimized}.")
        self._initialize(('VNS', optimized))
        time0 = time()
        if initial_tree is None:
            bushiness = (nber_scenarios,) + (1,) * (self.nber_stages-2)
            initial_tree = ScenarioTree.from_bushiness(bushiness)
            initial_tree.fill(self.scenario_process, optimized, self.variability_process, self.demerit)
            
        self._best_fod[('VNS', optimized)] = initial_tree.get_figure_of_demerit(self.demerit)
        self._best_tree[('VNS', optimized)] = copy.deepcopy(initial_tree)
        
        if initial_tree.depth <= 2:
            return self._best_tree[('VNS', optimized)], self._best_fod[('VNS', optimized)]
    
        iteration, no_improvement_count = 0, 0
        while iteration < max_iteration:
            try:
                iteration += 1
                nbreed = 1 #max(1, int(nber_scenarios / np.log(3*iteration)**neighborhood_shrink))
                candidates = [copy.deepcopy(self._best_tree[('VNS', optimized)]) for i in range(num_local_samples)]
    
                # increase neighborhood distance until improvement
                for neighborhood in range(1, num_neighborhoods + 1):
                    improved = False
        
                    # try multiple samples in the same neighborhood
                    for current_tree in candidates:
                        # increase neighborhood of current candidate
                        for ibreed in range(nbreed):
                            TreeSearch._tree_breed(current_tree) # split or merge
        
                        current_tree.fill(self.scenario_process, optimized, self.variability_process, self.demerit)
                        current_fod = current_tree.get_figure_of_demerit(self.demerit)
                        self._fod_sample[('VNS', optimized)].append(current_fod)
                        
                        if current_fod < self._best_fod[('VNS', optimized)]:
                            improved = True
                            self._best_tree[('VNS', optimized)] = current_tree
                            self._best_fod[('VNS', optimized)] = current_fod
                        
                    # start over if at least one sample provided improvement
                    if improved:
                        no_improvement_count = 0
                        break
        
                if not improved:
                    no_improvement_count += 1
                    if no_improvement_count >= max_no_improvement:
                        break
                    
                print(f"\riteration: {iteration}  demerit: {current_fod:.5f}  "
                      f"best demerit: {self._best_fod[('VNS', optimized)]:.5f}  "
                       f"no improvement count: {no_improvement_count}", end="")  
                
            except KeyboardInterrupt:
                break

        time1 = time()
        print(f"\nTotal number of iterations : {iteration} ({time1-time0:.1f} sec)")

    @staticmethod
    def _merge_nodes(node1, node2):
        assert node2 in node1.parent.children
        node1.parent.children.remove(node2)
        node1.add(*node2.children)

    @staticmethod
    def _split_node(node1, num_children_array):
        nodes = [Node() for nc in num_children_array]
        # redistribute subtrees
        for n, num_children in zip(nodes, num_children_array):
            for nc in node1.children[:num_children]:
                node1.children.remove(nc)
                n.add(nc)
        # remove old node to parent
        node1.parent.children.remove(node1)
        # add new nodes to parent
        node1.parent.add(*nodes)
        
    @staticmethod
    def _tree_breed(tree):
        # randomly pick a valid action
        actions = ['merge']

        splittable_nodes = [node for node in tree.nodes if node.parent and len(node.children) >= 2]
        if len(splittable_nodes) > 0:
            actions.append('split')

        mergeable = [node for node in tree.nodes if not node.is_leaf and node.has_siblings]
        action = 'split' if len(mergeable) == 0 else np.random.choice(actions)

        if action == 'merge':
            node1 = np.random.choice(mergeable)
            node2 = np.random.choice([n for n in node1.parent.children if n is not node1])
            TreeSearch._merge_nodes(node1, node2)

        elif action == 'split':
            node1 = np.random.choice(splittable_nodes)
            k = np.random.randint(len(node1.children) - 1) + 1
            num_children_array = [k, len(node1.children) - k]
            TreeSearch._split_node(node1, num_children_array)
        else:
            raise ValueError("unknown action")
            
    def exhaustive_search(self, 
                          nber_scenarios, 
                          optimized='forward',
                          min_branching_factor=1, 
                          max_iteration=np.inf):
        """Explore exhaustively the space of tree structures to find the scenario tree of lowest demerit"""
        assert optimized in ["forward", "backward"], ("`optimized` must be either 'forward' or "
        f"'backward', not {optimized}.")
        self._initialize(('EXH', optimized))
        time0 = time()
        iteration_count, no_improvement_count = 1, 0

        for structure in TreeSearch._exhaustive_structures(self.nber_stages, nber_scenarios, min_branching_factor):   
            try:
                current_tree = ScenarioTree(structure)
                current_tree.fill(self.scenario_process, optimized, self.variability_process, self.demerit)
                
                current_fod = current_tree.get_figure_of_demerit(self.demerit)
                self._fod_sample[('EXH', optimized)].append(current_fod)
                
                if current_fod < self._best_fod[('EXH', optimized)]:
                    improved = True
                    self._best_tree[('EXH', optimized)] = current_tree
                    self._best_fod[('EXH', optimized)] = current_fod
                else:
                    improved = False
                    
                no_improvement_count += 1 if not improved else 0
                    
                if iteration_count % 10 == 0:
                    print(f"\riteration: {iteration_count}  demerit: {current_fod:.5f}  "
                          f"best demerit: {self._best_fod[('EXH', optimized)]:.5f}  "
                           f"no improvement count: {no_improvement_count}", end="")        
            except KeyboardInterrupt:
                break
            
            iteration_count += 1
            if iteration_count > max_iteration:
                break

        time1 = time()
        print(f"\nTotal number of iterations : {iteration_count-1} ({time1-time0:.1f} sec)")
        
    @staticmethod
    def _exhaustive_structures(depth, N, b):
        """Generates all tree structures with depth T, N scenarios, and a branching lowerbound b""" 
        if depth == 2:
            yield Node.from_bushiness((N,))
            return

        for n in range(b**(depth-1), int(N / b) + 1):            
            for tree in TreeSearch._exhaustive_structures(depth-1, n, b):                
                for new_tree in TreeSearch._extend_structure(tree, N, b):        
                    new_tree.delete_data(["pos", "n"])
                    yield new_tree
                         
    @staticmethod        
    def _pseudo_integer_partitions(cardinality, integer, lowerbounds):
        """Enumeration of all integer partitions (not permutation free) of a fixed cardinality. 
        The partition of an integer k is a tuple (n_1, ..., n_m) such that:
            (i) k = n_1 + ... + n_m
            (ii) n_i is integer >= 1.
            
        Unlike the regular integer partitions (method ._integer_partitions()), here the condition:
            n_1 <= n_2 <= ... <= n_m 
        need not be satisfied.
        
        The lowerbounds adds the condition:
            (iv) n_i >= lowerbound[i], for i = 1, ...,m (componentwise lowerbound).
        
        Arguments:
        ----------
        cardinality: integer >= 1
            The number of elements partitionning the integer.
            
        integer: integer >= 1
            The integer being partitionned.
            
        lowerbounds: tuple of integers >= 1
            The componentwise lowerbound on the elements partitionning the integer.
            
        Returns:
        --------
        iterator on the partitions.
        """
        if cardinality == 1:
            if integer >= lowerbounds[0]:
                yield (integer,)
                return
        for i in range(lowerbounds[0], integer-sum(lowerbounds[1:])+1):
            for t in TreeSearch._pseudo_integer_partitions(cardinality-1, integer-i, lowerbounds[1:]):
                yield (i,) + t
    
    @staticmethod               
    def _integer_partitions(cardinality, integer, lowerbound=1):
        """Enumeration of all integer partitions of a fixed cardinality. A partition of cardinality m of 
        an integer k is a tuple (n_1, ..., n_m) such that:
            (i) k = n_1 + ... + n_m
            (ii) n_i is integer >= 1
            (iii) n_1 <= n_2 <= ... <= n_m.
            
        The lowerbound (optional) adds the condition:
            (iv) n_i >= lowerbound, for i = 1, ...,m.
        
        Arguments:
        ----------
        cardinality: integer >= 1
            The number of elements partitionning the integer.
            
        integer: integer >= 1
            The integer being partitionned.
            
        lowerbound: integer >= 1 (default 1)
            The lowerbound on the elements partitionning the integer.
            
        Returns:
        --------
        iterator on the partitions.
        """
        if cardinality == 1:
            if integer >= lowerbound:
                yield (integer,)
                return
        for i in range(lowerbound, integer-(cardinality-1)*lowerbound+1):
            for t in TreeSearch._integer_partitions(cardinality-1, integer-i, i):
                yield (i,) + t
    
    @staticmethod
    def _extend_structure(tree, N, b):
        """This generator takes any tree structure and generates all 
        tree structure with N scenarios and 1 stage more."""
        partitionP1, partitionP2, P2, lowerboundsP2  = {}, {}, {}, {}
    
        #Create the Partition P1 of leaf nodes
        for leaf in tree.leaves:
            history = tuple([len(n.children) for n in leaf.branch if n != leaf])
        
            if history in partitionP1.keys():
                partitionP1[history] += [leaf] 
            else:
                partitionP1[history] = [leaf] 
            P1 = len(partitionP1.keys())
    
        lowerboundsP1 = [b * len(partitionP1[key]) for key in partitionP1.keys()]
        
        #Create the Partition P2 of leaf nodes
        for i, key in enumerate(partitionP1.keys()):
            partitionP2[key] = list(set([tuple(leaf.parent.children) for leaf in partitionP1[key]]))
            P2[key] = len(partitionP2[key])
            lowerboundsP2[key] = len(partitionP2[key][0])
    
        #Create a data 'pos' (a 3-tuple) to index each leaf in its partition and subpartition
        for i, key in enumerate(partitionP1.keys()):
            for j, leaves in enumerate(partitionP2[key]):
                for k, leaf in enumerate(leaves):
                    leaf.data["pos"] = (i, j, k)
           
        #generate the integer tuples x, y, z to code the tree branching
        indexing_A = TreeSearch._pseudo_integer_partitions(P1, N, lowerboundsP1)
        for x in indexing_A: 
            indexing_B = lambda i, key: list(TreeSearch._integer_partitions(P2[key], 
                                                                                 x[i], 
                                                                                 b*lowerboundsP2[key]))

            for y in Cartesian(*[indexing_B(i, key) for i, key in enumerate(partitionP1.keys())]):
    
                set_B = lambda i, j, key: list(TreeSearch._integer_partitions(lowerboundsP2[key], 
                                                                                   y[i][j], 
                                                                                   b))
                for z in Cartesian(*[Cartesian(*[set_B(i, j, key) for j in range(P2[key])]) 
                                                                for i, key in enumerate(partitionP1.keys())]):
                    new_tree = copy.deepcopy(tree)
                    
                    for leaf in list(new_tree.leaves):
                        (i, j, k) = leaf.data["pos"]
                        leaf.add(*[Node() for i in range(z[i][j][k])])
    
                    yield new_tree
    
                    
                    


# -*- coding: utf-8 -*-

import numpy as np

from ..uncertainty_basis import UncertaintyBasis

class ClientsPresence(UncertaintyBasis):
    
    def __init__(self, n_scenarios, n_client_locations, p=None, lb=None, ub=None):
        self.n_client_locations = n_client_locations
        self.n_scenarios = n_scenarios
        
        self.p = p
        self.lb = lb
        self.ub = ub
        self._scenarios = None
        
        if self.p is None and self.lb is None and self.ub is None:
            raise TypeError("Provide p or (lb, ub)")
        elif self.p is not None and self.lb is not None and self.ub is not None:
            raise TypeError("Provide p or (lb, ub) but not both")
        
        self.randomize()
        UncertaintyBasis.__init__(self, 
                                  name=f"Presence uncertainty",
                                  n_scenarios=self.n_scenarios,
                                  features={'h': self.n_client_locations})
        
    def randomize(self):
        self._scenarios = np.array([self.bernoulli() for _ in range(self.n_scenarios)])
        
    def get_scenario(self, scen_index):
        assert isinstance(scen_index, int), f"`scen_index should be of type int, not {type(scen_index)}."
        return {'h': self._scenarios[scen_index]}
    
    def bernoulli(self):
        if self.p is not None:
            return np.random.binomial(n=1, p=self.p, size=self.n_client_locations)
        else:
            p = np.random.uniform(self.lb, self.ub)
            return np.random.binomial(n=1, p=p, size=self.n_client_locations)


# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
import copy

from typing import Union, List, Dict, Tuple, Optional, Callable, Iterator, Any
import numpy as np

from stochoptim.scengen.scenario_tree import ScenarioTree, decompose
from stochoptim.scengen.tree_structure import Node
from stochoptim.scengen.decision_process import DecisionProcess
from stochoptim.stochprob.stochastic_solution_basis import StochasticSolutionBasis
from stochoptim.util import TimeIt

Num = Union[int, float]
Subscript = Union[int, str, Tuple[Union[int, str], ...]]

class StochasticProblemBasis(ABC):
    
    def __init__(self, 
                 name: str,
                 n_stages: int,
                 objective_sense: str,
                 is_obj_random: bool, 
                 is_mip: bool,
                 solution_class: Optional[StochasticSolutionBasis] = None,
                 initialize_indices: bool = True):
        
        self._name = name
        self._n_stages = n_stages
        self._objective_sense = objective_sense
        self._is_obj_random = is_obj_random
        self._is_mip = is_mip
        
        # class of the stochastic solution
        if solution_class is not None:
            self._solution_class = solution_class
        else:
            self._solution_class = StochasticSolutionBasis
            
        # variables indices
        # map: stage -> var_name -> var_subscript -> index
        self._map_dvar_to_index = {}
        self._map_rvar_to_index = {}
        # map: (stage, index) -> (var_name, var_subscript)
        self._map_index_to_dvar = {}
        self._map_index_to_rvar = {} 
        # map: stage -> var_name -> (var_type, lb, ub)
        self._map_stage_to_var_type = {}
            
        # precomputation
        # map: stage -> list of var_names
        self._name_precomputed_decisions = {}
        self._name_precomputed_parameters = {}        
        
        # path information 
        self._node = None
        self._decision_path = None
        self._scenario_path = None
        self._node_address = None
        self._memory_path = None
        self._sum = None
        self._dot = None
                    
        if initialize_indices:
            self._initialize_indices()
        self._initialize_precomputation()
        
    # --- Property ---   
    @property
    def name(self):
        return self._name
    
    @property
    def n_stages(self):
        return self._n_stages
    
    @property
    def objective_sense(self):
        return self._objective_sense
    
    @property
    def is_obj_random(self):
        return self._is_obj_random
    
    @property
    def is_mip(self):
        return self._is_mip
    
    @property
    def map_stage_to_var_type(self) -> Dict[int, Dict[str, Tuple[str, Optional[Num], Optional[Num]]]]:
        return self._map_stage_to_var_type
    
    @property
    def map_dvar_to_index(self) -> Dict[int, Dict[str, Dict[Subscript, int]]]:
        return self._map_dvar_to_index
    
    @property
    def map_dvar_name_to_nb(self) -> Dict[int, Dict[str, int]]:
        return {stage: {var_name: len(variables) for var_name, variables in dvar.items()} 
                    for stage, dvar in self._map_dvar_to_index.items()}
    
    @property
    def map_stage_to_dvar_names(self) -> Dict[int, List[str]]:
        return {stage: list(dvar.keys()) for stage, dvar in self._map_dvar_to_index.items()}
    
    @property
    def map_rvar_to_index(self) -> Dict[int, Dict[str, Dict[Subscript, int]]]:
        return self._map_rvar_to_index
    
    @property
    def map_rvar_name_to_nb(self) -> Dict[int, Dict[str, int]]:
        return {stage: {var_name: len(variables) for var_name, variables in rvar.items()} 
                    for stage, rvar in self._map_rvar_to_index.items()}
    
    @property
    def map_stage_to_rvar_names(self) -> Dict[int, List[str]]:
        return {stage: list(rvar.keys()) for stage, rvar in self._map_rvar_to_index.items()}
    
    @property       
    def map_index_to_dvar(self) -> Dict[Tuple[int, str, int], Subscript]:
        return self._map_index_to_dvar
    
    @property       
    def map_index_to_rvar(self) -> Dict[Tuple[int, str, int], Subscript]:
        return self._map_index_to_rvar
    
    @property
    def name_precomputed_decisions(self) -> Dict[int, List[str]]:
        return self._name_precomputed_decisions
    
    @property
    def name_precomputed_parameters(self) -> Dict[int, List[str]]:
        return self._name_precomputed_parameters
    
    # --- Abstract methods ---
    @abstractmethod
    def decision_variables_definition(self, stage):
        """For each decision variables of type {y_{t, i} \in A : i \in I} where I is an indexing set 
        (a list of tuples), y_{t,i} is of type A (with A = binary (B), or integer (I), or continuous (C)), 
        and lb <= y_{t,i} <= ub, this function generates a 5-tuple of the form: ('y', I, lb, ub, A) 
        under the conditon stage == t. (lb and ub can be two iterables provided they match same size of I)"""
        pass
    
    @abstractmethod
    def random_variables_definition(self, stage):
        """For each random variable of type {xi_{stage, i} : i \in I} where I is an indexing set (a list of tuples), 
        this method generates a 2-tuple of the form ('xi', I)"""
        pass
    
    @abstractmethod
    def deterministic_linear_constraints(self, stage):
        pass
    
    @abstractmethod
    def random_linear_constraints(self, stage):
        pass
    
    def precompute_decision_variables(self, stage: int) -> Iterator[Tuple[str, Callable[[], Any]]]:
        pass
    
    def precompute_parameters(self, stage: int) -> Iterator[Tuple[str, Callable[[], Any]]]:
        pass
    
    def sanity_check(self, stage):
        pass
    
    def objective(self):
        pass
    
    def objective_by_stage(self):
        pass
    
    def deterministic_indicator_constraints(self, stage):
        pass
    
    def random_indicator_constraints(self, stage):
        pass
    
    def deterministic_equivalence_constraints(self, stage):
        pass
    
    def random_equivalence_constraints(self, stage):
        pass
    
    def random_quadratic_constraints(self, stage):
        pass
    
    def deterministic_quadratic_constraints(self, stage):
        pass
    
    def deterministic_if_then_constraints(self, stage):
        pass
    
    def random_if_then_constraints(self, stage):
        pass
    
    # --- To build approximate problem ---      
    def sum(self, args):
        if isinstance(args, list):
            args = np.array(args)
        if args.dtype == np.dtype('O'):
            return self._sum(args)
        else:
            return np.sum(args)
    
    def dot(self, var, coef):
        """
        var: list or 1d-array
        
        coef: int, float, list or 1d-array
        """
        if isinstance(var, list):
            var = np.array(var)
        # check that coef is a number of an array of the same size as var
        assert np.array(coef).shape == () or np.array(coef).shape[0] == var.shape[0], \
                f"The length of the arrays of variables ({var.shape[0]}) and coefficients ({len(coef)}) do not match."
        if var.dtype == np.dtype('O'): # instance of Var, LinearExpr
            coef = coef.tolist() if isinstance(coef, np.ndarray) else coef # docplex .dot is faster on a list
            return self._dot(var, coef)
        else:
            # if coef is a number we repeat it in an array otherwise np.dot behaves as np.multiply
            if np.array(coef).shape == (): 
                coef = coef * np.ones(var.shape[0]) # faster than np.tile
            return np.dot(var, coef)

    def set_path_info(self, node, model=None):
        """Set the decision path and scenario path (and others) that will be used in building the approximate 
        problem."""
        self._node = node
        self._decision_path = Node.get_data_path(node, 'decision', default={})
        self._scenario_path = ScenarioTree.get_scenario_path(node)
        self._node_address = node.address
        self._memory_path = Node.get_data_path(node, 'memory', default={})
        self._sum = np.sum if model is None else model.sum
        self._dot = np.dot if model is None else model.dot
        self._constraints_generators = \
            {('deterministic', 'linear'): self.deterministic_linear_constraints(node.level),
             ('deterministic', 'indicator'): self.deterministic_indicator_constraints(node.level),
             ('deterministic', 'quadratic'): self.deterministic_quadratic_constraints(node.level),
             ('random', 'linear'): self.random_linear_constraints(node.level),
             ('random', 'indicator'): self.random_indicator_constraints(node.level),
             ('random', 'quadratic'): self.random_quadratic_constraints(node.level)}
            
    def average_over_distribution(self, expression: Callable[[], Any], stage: int):
        """Average an expression over all the nodes at a given stage that are in the subtree of
        the underlying node from which the function is called (a.k.a. `self._node`)"""
        avg = 0
        for n in self._node.nodes_at_level(stage):
            self._decision_path = Node.get_data_path(n, 'decision', default={})
            self._scenario_path = ScenarioTree.get_scenario_path(n)
            self._memory_path = Node.get_data_path(n, 'memory', default={})
            self._node_address = n.address
            avg += (n.data['W'] / self._node.data['W']) * expression()
        self.set_path_info(self._node)
        return avg
    
    def generate_node_constraints(self, node, ct_type, model=None, remove_constraints=None):
        """
        ct_type: tuple (a, b) where a is 'deterministic' or 'random' and b is 'linear' or 'indicator'.
        
        remove_constraints: list of str or None (default: None)
            Contains the name of the constraints to be removed from the problem.
            If None, no constraint is removed.
        """
        self.set_path_info(node, model)   
        if self._node.data.get(f'{ct_type[0]}-ct') is None:
            self._node.data[f'{ct_type[0]}-ct'] = [] # will contain additionnal linear cts
            
        if not hasattr(self._constraints_generators[ct_type], '__next__'): # if not an iterator
            return iter(()) # empty iterator
        
        for ct_generator in self._constraints_generators[ct_type]:
            assert hasattr(ct_generator, '__next__'), ("Constraints should be given as an "
            f"iterator, it was of type {type(ct_generator)} at node {node.address}.")
            if remove_constraints is not None and ct_generator.__name__ in remove_constraints:
                continue
            for ct_tuple in ct_generator:
                if ct_type[1] == 'linear':
                    new_ct_tuple = self._inspect_linear_constraint(ct_tuple)
                elif ct_type[1] == 'indicator':
                    new_ct_tuple = self._inspect_indicator_constraint(ct_tuple, ct_type)
                elif ct_type[1] == 'equivalence':
                    raise NotImplementedError()
                elif ct_type[1] == 'if-then':
                    raise NotImplementedError()
                elif ct_type[1] == 'quadratic':
                    new_ct_tuple = ct_tuple
                else:
                    raise ValueError("Wrong type of constraints, must be 'linear', 'indicator', 'equivalence' or "
                                     f"'if-then', not {ct_type[1]}")
                if new_ct_tuple is not None:
                    yield new_ct_tuple
                
    def objective_path(self, leaf, model=None):
        self.set_path_info(leaf, model)
        return self.objective()

    def objective_node(self, node, model=None):
        self.set_path_info(node, model)
        return self.objective_by_stage(node.level)
    
    # --- check constraints validity ---
    def _inspect_linear_constraint(self, ct_tuple):
        # check the tuple input
        if not isinstance(ct_tuple, tuple):
            ct_tuple = (ct_tuple, '') # append empty name to it
        else:
            assert len(ct_tuple) == 2, \
            f"Each linear constraint should be a 2-tuple (constraint, name), instead we received type {type(ct_tuple)}."
        # check specifically the constraint    
        ct, name = ct_tuple
        if self.is_constraint_relevant(ct, name):
            return ct_tuple
        else:
            assert self.is_constraint_satisfied(ct, name), \
                    f"Constraint '{name}' is not satified at node address {self._node_address}"
       
    def _inspect_indicator_constraint(self, ct_tuple, ct_type):
        # check the tuple input
        if len(ct_tuple) == 3:
            ct_tuple = ct_tuple + ('_',)
        elif len(ct_tuple) != 4:
            raise ValueError("Each indicator constraint should be a 3- or 4-tuple (indicator var, constraint, "
                             f"activation value, optional: name), instead we received {ct_tuple}")
        # check specifically the constraint and indicator variable
        var, ct, true_value, name = ct_tuple
        if np.array(var).dtype == np.dtype('O'): # instance of docplex Var (check if indicator variable is not fixed beforehand)
            if self.is_constraint_relevant(ct, name):
                assert var.is_binary(), \
                    f"The activation variable '{var}' in the indicator constraint '{name}' must be binary."
                return ct_tuple
            else:
                if not self.is_constraint_satisfied(ct, name):
                    self._node.data[f'{ct_type[0]}-ct'] += [(var == int(not true_value), name)]
        else: # if indicator variable is fixed beforehand
            if var == true_value:
                if self.is_constraint_relevant(ct, name):
                    self._node.data[f'{ct_type[0]}-ct'] += [(ct, name)]
                else:
                    assert self.is_constraint_satisfied(ct, name), (f"Indicator constraint {name} is not satisfied "
                                                                    f"at node address {self._node_address}")
            
    def is_constraint_relevant(self, ct, name):
        """Return True if the constraint includes a least one decision variable.
        Note: a constraint with no decision variables (i.e., with both side constant) may arise when some of the 
        decisions are fixed beforehand."""
        if np.array(ct).dtype == np.dtype('O'): # if the constraint is a docplex linear expression
            if ct.right_expr.is_constant() and ct.left_expr.is_constant(): # if left-hand side and right-hand side are both constant
                return False
            else:
                return True
        elif isinstance(ct, (bool, np.bool_, np.bool)): # if the constraint is a boolean
            return False
        else:
            raise TypeError(f"Unknown type for constraint '{name}' at node address {self._node_address}: {type(ct)}")
                    
    def is_constraint_satisfied(self, ct, name):
        """Return True if a constraint with both sides constant evaluates as True.
        Note: a constraint with both side constant may arise when some of the decisions are fixed beforehand."""
        if np.array(ct).dtype == np.dtype('O'): # if the constraint is a docplex linear expression
            return eval(ct.to_string())
        elif isinstance(ct, (bool, np.bool_, np.bool)):
            return ct
        else:
            raise TypeError(f"Unknown type for constraint '{name}' at node address {self._node_address}: {type(ct)}")
            
    # --- Define and access variables and indices ---
    def get_dvar(self, 
                 stage: int, 
                 var_name: str, 
                 var_subscripts: Optional[Subscript] = None) -> Union[Num, np.ndarray]:
        """ Access a decision variable. 
        This parent method is the one that should be used in defining a stochastic problem in a child class of 
        StochasticProblemBasis.
        
        Note: the method `.set_path_info` must be called before calling this method.
        
        Arguments:
        ----------
        stage: int >= 0
        
        var_name: str
        
        var_subscripts: str, int, or tuple of int/str
            The appropriate type is the one used to define the variables in the .decision_variables_definition().
            
        Returns:
        --------
        int/float or 1d-array: the decision variable(s).
        """
        try:
            decisions = self._decision_path[stage][var_name]
        except KeyError:
            if self._decision_path.get(stage) is None:
                raise KeyError(f"Stage {stage} is not valid at node of address {self._node_address}. "
                               f"Stage should be <= {len(self._node_address)}")
            elif self._decision_path[stage].get(var_name) is None:
                raise KeyError(f"Variable '{var_name}' is not in the decision process at stage "
                               f"{stage}. Variables in the decision process are: {list(self._decision_path[stage].keys())}")   
        if var_subscripts is None:
            return decisions
        else:
            index = self.get_dvar_index(stage, var_name, var_subscripts)
            try:
                return decisions[index]
            except IndexError:
                raise IndexError(f"Variable subscript {var_subscripts} (index = {index}) is out of bounds for '{var_name}' "
                f"at stage {stage}, which has {decisions.shape[0]} elements: {decisions}")
                
    def get_rvar(self, 
                 stage: int, 
                 var_name: str, 
                 var_subscripts: Optional[Subscript] = None) -> Union[Num, np.ndarray]:
        """ Access a random variable. 
        This parent method is the one that should be used in defining a stochastic problem in a child class of 
        StochasticProblemBasis.
        
        Note: the method `.set_path_info` must be called before calling this method.
        
        Arguments:
        ----------
        stage: int >= 1
        
        var_name: str
        
        var_subscripts: str, int, or tuple of int/str
            The appropriate type is the one used to define the variables in the .random_variables_definition().
        
        Returns:
        --------
        int/float or 1d-array: the random variable(s).
        """
        assert var_name in self.map_stage_to_rvar_names[stage], \
            (f"Random variable '{var_name}' is not in the list of existing "
             f"variables at stage {stage}: {self.map_stage_to_rvar_names[stage]}.")
        assert self._scenario_path is not None, \
            "The method `set_path_info(node)` should be called in order to access a random variable at a node."
      #  assert stage <= len(self._node_address), (f"Information on variable '{var_name}' at stage {stage} "
      #                     f"is not available at the level {len(self._node_address)} in the scenario")
        if var_subscripts is None:
            return self._scenario_path[stage][var_name]
        else:
            index = self.get_rvar_index(stage, var_name, var_subscripts)
            assert index <= self._scenario_path[stage][var_name].shape[0] - 1, \
                (f"Variable subscript {var_subscripts} (index = {index}) is out of bounds for '{var_name}' "
                f"at stage {stage}, which has {self._scenario_path[stage][var_name].shape[0]} elements: "
                f"{self._scenario_path[stage][var_name]}")
            return self._scenario_path[stage][var_name][index]
                
    def get_memory(self, stage, var_name):
        """Return the value stored in the scenario tree data memory"""
        assert self._memory_path is not None, \
            "The method `set_path_info(node)` should be called in order to access memory information at a node."
        return self._memory_path.get(stage, {}).get(var_name)
    
    def get_dvar_index(self, stage, var_name, var_subscripts):
        """Return the index of a decision variable."""
        return self._map_dvar_to_index[stage][var_name][var_subscripts]
        
    def get_rvar_index(self, stage, var_name, var_subscripts):
        """Return the index of a random variable."""
        return self._map_rvar_to_index[stage][var_name][var_subscripts]
        
    def get_dvar_from_index(self, stage, var_name, index):
        return self._map_index_to_dvar[stage, var_name, index]
    
    def get_rvar_from_index(self, stage, var_name, index):
        return self._map_index_to_rvar[stage, var_name, index]

    def _initialize_indices(self):
        """Set the mapping from the decision variables and random parameters to their corresponding indices, 
        and vice versa"""        
        for stage in range(self._n_stages):                        
            # decision variables
            self._map_dvar_to_index[stage] = {}
            self._map_stage_to_var_type[stage] = {}
            for var_name, indexing_set, lb, ub, var_type in self.decision_variables_definition(stage):
                if len(indexing_set) == 0:
                    continue
                # get variable type and lb, ub
                self._map_stage_to_var_type[stage][var_name] = (var_type, lb, ub)
                # get indices
                self._map_dvar_to_index[stage][var_name] = {}
                for index, var_subscripts in enumerate(indexing_set):
                    self._map_dvar_to_index[stage][var_name][var_subscripts] = index
                    self._map_index_to_dvar[stage, var_name, index] = var_subscripts
                    
            # random variables
            self._map_rvar_to_index[stage] = {}
            for var_name, indexing_set in self.random_variables_definition(stage):
                if len(indexing_set) == 0:
                    continue
                # get indices
                self._map_rvar_to_index[stage][var_name] = {}
                for index, var_subscripts in enumerate(indexing_set):
                    self._map_rvar_to_index[stage][var_name][var_subscripts] = index
                    self._map_index_to_rvar[stage, var_name, index] = var_subscripts
               
            var_names_in_common = set(self._map_dvar_to_index[stage].keys())\
                                    .intersection(set(self._map_rvar_to_index[stage].keys()))
            assert var_names_in_common == set(), ("Decision variables and random variables have the same name: "
                                                    f"{var_names_in_common} at stage {stage}.")
             
    # --- Solve methods ---
    def solve(self,
              *scenario_trees: ScenarioTree,
              decision_process: Optional[DecisionProcess] = None,
              decomposition: bool = True,
              precompute_parameters: bool = False,
              precompute_decisions: bool = False,
              relaxation: bool = False,
              keep_integer: Optional[Dict[int, List[str]]] = None,
              remove_constraints: Optional[List[str]] = None,
              warmstart: Optional[Dict[str, Union[float, int]]] = None,
              find_only_feasibility: bool = False,
              with_variable_name: bool = False,
              objective_type: str = 'path',
              verbose: int = 3,
              mip_filename: Optional[str] = None,
              solve_problem: bool = True,
              tree_filename: Optional[str] = None,
              tree_extension: str = 'txt',
              with_keys: Optional[List[str]] = None,
              without_keys: Optional[List[str]] = None,
              **kwargs) -> Union[StochasticSolutionBasis, List[StochasticSolutionBasis]]:
        """ Main method to solve the stochastic problem on one or several scenario trees.
        
        Arguments:
        ----------
        scenario_trees: instances of ScenarioTree
            The scenario trees for which the stochastic problem will be solved.
            
        decision_process: instance of DecisionProcess
            The decisions variables that are fixed to some value beforehand.
            If None, no decisions are fixed.
            
        decomposition: bool (default: True)
        
        precompute_parameters: bool (default: True)
        
        precompute_decisions: bool (default: True)
        
        relaxation: bool (default: False): 
            If True, the optimizer solves the continuous relaxation of the problem.
         
        keep_integer: Dict[int, List[str]] or None (default: None)
            All the variable names that must stay integer even if the problem is relaxed.
            
        remove_constraints: List[str] or None (default: None)
            Contains the name of the constraints to be removed from the problem.
            If None, no constraint is removed.
            
        warmstart: Dict[str, Union[float, int]] or None (default: None)        
           
        find_only_feasibility: bool (default: False)
            If True, the solver will only look for a feasible solution and not an optimal one.
            
        with_variable_name: bool (default: False)
            If True, the model will refer to the variables by their own names as given in the problem's definition.
            
        objective_type: {'path', 'node'} (default: 'path')
        
        verbose: {0, 1, 2, 3} (default: 3)
                
        mip_filename: str or None (default: None)
            If a str is given, it is the path where the problem will be saved.
            The path should contain the extension '.lp' or '.sav'.
            
        solve_problem: bool (default: True)
            If True, solves the stochastic problem. 
            If False, only builds the problem.
            
        kwargs:
        -------
        check_sanity: bool (default: False)
            If True, execute the sanity check specified in the stochastic problem.
        
        check_fixed_constraints: bool (default: False)
            If True, check that the constraints with decisions fixed beforehand in the decision process are satisfied. 
            Note that this may be time consuming.
            
        check_topology: bool (default: True)
            If True, it is checked whether all scenario trees have equal topology (which is a necessary condition
            for them to be solved together).
                 
        fill_scenario_tree: bool (default: True)
            If False, only the solve details at the root and subroots will be available (solve details are 
            the objective value, the best bound, the solve status and the solving time).
                                   
        clear_between_trees: bool (default: True)
            If True, the solution of one scenario tree is not used as warmstart for the next ones.
                        
        timelimit: int >= 1 or None (default: None)
            Time limit for the optimization of one stochastic problem. If None, no limit is set.

        logfile: str or None (default: "")
            This string is appended to the log filename. This file is located in a folder 'logfiles' in the 
            current directory (the folder is created if it doesn't already exist). If None, no log is available. 
            
        Returns:
        --------
        An instance or a list of instances of the class `self._solution_class`.
        `self._solution_class` is either the class StochasticSolutionBasis or a subclass
        """ 
        from stochoptim.scengen.approx_prob import ApproximateProblem, _timeit

        # check scenario trees
        self._check_scenario_trees(scenario_trees)
        
        # decision process
        if decision_process is None:
            decision_process = DecisionProcess(self._map_dvar_to_index)
        else:
            assert self.map_dvar_to_index == decision_process.map_dvar_to_index, \
             "The decision variables in the decision process do not match those of the stochastic problem."
             
        # check file names
        mip_filename = self._check_filenames(mip_filename, 'mip_filename', len(scenario_trees))
        tree_filename = self._check_filenames(tree_filename, 'tree_filename', len(scenario_trees))
            
        if verbose >= 1:
            print(f"Number of scenario trees: {len(scenario_trees)} (bushiness: {scenario_trees[0].bushiness}) \n")

        # build and solve approximate problem  (trees are copied to store memory data at the nodes)
        approximate_problem = ApproximateProblem(scenario_trees[0].copy(),
                                                 stochastic_problem=copy.copy(self),
                                                 decision_process=decision_process, 
                                                 decomposition=decomposition, 
                                                 precompute_parameters=precompute_parameters,
                                                 precompute_decisions=precompute_decisions,
                                                 relaxation=relaxation, 
                                                 keep_integer=keep_integer,
                                                 remove_constraints=remove_constraints,
                                                 warmstart=warmstart,
                                                 find_only_feasibility=find_only_feasibility,
                                                 objective_type=objective_type,
                                                 verbose=verbose,
                                                 **kwargs)
        
        solutions = [None] * len(scenario_trees)
        approximate_problem.build_deterministic_part()
        # solve iteratively over the trees
        for index_tree in range(len(scenario_trees)):
            with _timeit(approximate_problem, 
                         f"\rSolve scenario tree #{index_tree+1}... " if verbose <= 2 \
                         else f"\nSolve scenario tree #{index_tree+1}... \n", 1):
                # build, solve, and fill        
                approximate_problem.build_random_part()
                if mip_filename:
                    approximate_problem.write_docplex_models(mip_filename[index_tree])
                if solve_problem:
                    approximate_problem.solve_docplex_models()
                    approximate_problem.fill_decisions_in_tree(tree_filename[index_tree], tree_extension,
                                                               with_keys, without_keys)
                    # append solution
                    solutions[index_tree] = self._solution_class(copy.copy(self), 
                                                                 approximate_problem.scenario_tree)
                # set next scenario tree
                if index_tree < len(scenario_trees)-1:
                    approximate_problem.scenario_tree = scenario_trees[index_tree+1].copy()
                    approximate_problem.scenario_tree.append_data_dict(approximate_problem.dvar_dict)
                    approximate_problem.remove_random_constraints()

        return solutions[0] if len(scenario_trees) == 1 else solutions
            
    def compute_wait_and_see(self, 
                             scenario_tree: ScenarioTree, 
                             **kwargs) -> List[StochasticSolutionBasis]:
        """ 
        Solve the Wait-and-See (WS) problems. Each WS problem is formulated by taking a single scenario in 
        the scenario tree. 
        
        Parameters:
        -----------
        scenario_tree: instance of ScenarioTree
            The scenario tree representing the stochastic problem uncertainty.
            
        kwargs: 
        -------
        see `solve` method
        """
        return self.solve(*decompose(scenario_tree), **kwargs)
        
    def compute_opportunity_cost(self, 
                                 scenario_tree: ScenarioTree, 
                                 return_matrix=False,
                                 relaxation_step1=False,
                                 relaxation_step2=False,
                                 timelimit_step1=None,
                                 timelimit_step2=None,
                                 fill_scenario_tree_step2=True,
                                 **kwargs) -> List[StochasticSolutionBasis]:            
        r"""
        Compute the square-matrix of opportunity cost. The computation consists in two steps:
            (1) solving the two-stage problem for each scenario independently;
            (2) solving the 2nd-stage problem for each scenario and each 1st-stage solution computed at step (1).
        
        Arguments:
        ----------
        scenario_tree: instance of ScenarioTree
        
        relaxation_step1: bool (optional)
            If True, the optimizer solves the continuous relaxation at step (1).
        
        relaxation_step2: bool (optional)
            If True, the optimizer solves the continuous relaxation at step (2).
            
        timelimit_step1: int >= 1 or None (default: None)
            Time limit for the optimization of one stochastic problem at step (1). If None, no limit is set.
            
        timelimit_step2: int >= 1 or None (default: None)
            Time limit for the optimization of one stochastic problemat step (2). If None, no limit is set.
            
        kwargs:
        -------
        all kwargs of `solve` method, except `decision_process`, `timelimit`, `verbose`, `relaxation`.
        
        Returns:
        --------
        List of instances of StochasticSolutionBasis of length `n_leaves` where n_leaves is the number of leaves 
        in the scenario tree. 
            The i-th element is the solution corresponding to the 2nd-stage evaluation of the 1st-stage decisions 
            of the i-th wait-and-see problem. Formally, it is: sum_{j=1,...,n_leaves} F(x_i^*, \xi_j) where x_i^* 
            is the 1st-stage solution of the i-th WS problem and \xi_j is the j-th scenario of the scenario tree.
        """
        # solve wait and see problems
        with TimeIt(f"Solve Wait-and-See Problems... ", 
                    lambda t: f"Finished {t:.3f} sec\n"):
            ws_solutions = self.compute_wait_and_see(scenario_tree, 
                                                     decision_process=None,
                                                     relaxation=relaxation_step1,
                                                     timelimit=timelimit_step1,
                                                     fill_scenario_tree=True,
                                                     **kwargs)
        # evaluate wait and see problems
        #matrix = np.zeros((len(ws_solutions), len(ws_solutions)))
        eval_ws_solution = []
        for k, ws_sol in enumerate(ws_solutions):
            with TimeIt(f"Evaluate Wait-and-See Solution #{k}... ", 
                        lambda t: f"Finished {t:.3f} sec\n"):
                eval_ws_solution.append(
                        self.solve(scenario_tree,
                                   decision_process=DecisionProcess(self._map_dvar_to_index, {0: ws_sol.x0}),
                                   relaxation=relaxation_step2,
                                   timelimit=timelimit_step2,
                                   fill_scenario_tree=fill_scenario_tree_step2,
                                   **kwargs))
        if return_matrix:
            return np.array([sol.objective_value_at_leaves for sol in eval_ws_solution])
        else:
            return ws_solutions, eval_ws_solution

    def compute_value_proxy(self, 
                            scenario_tree_true,
                            scenario_tree_proxy,
                            **kwargs):
        """ 
        Compute the Value of the Proxy Distribution (VPD).
        
        Note: it is defined as:
            VPD = v(EPD) - v(SP) (for mininization) or v(SP) - v(EPD) (for maximization)
        where:
            v(SP) is the optimal-value of the Two-Stage Problem (SP)
            v(EPD) is the optimal-value of the Expectation of Proxy Distribution Problem (EPD)
        The EPD problem evaluates the value of the Proxy Distribution (PD) solution in the true stochastic problem:
            v(EPD) = v(SP s.t. x0 = x0(PD))
        where x0(PD) is the stage-0 optimal solution obtained using the proxy scenarios.
           
        Arguments:
        ----------
        scenario_tree_true: instance of ScenarioTree
            The scenario tree of the true stochastic problem (SP).
            
        scenario_tree_proxy: instance of ScenarioTree
            The scenario tree of the proxy stochastic problem (PD).
                        
        kwargs: 
        -------
        all kwargs of `solve` method, except `decision_process`.
        
        Returns:
        --------
        dictionary: mapping:
            'PD': instance of StochasticSolutionBasis (the PD solution instance)
            'EPD': instance of StochasticSolutionBasis (the EPD solution instance)
            'SP':  instance of StochasticSolutionBasis (the SP solution instance)
            'VPD': float (the Value of Proxy Distribution)
            'VPD%' float (the Value of Proxy Distribution in percentage of v(2SP))
        """
        # Problem with proxy scenarios
        print("*** Solve Proxy Distribution Problem (PD) ***")
        proxy_dist_sol = self.solve(scenario_tree_proxy,
                                    decision_process=None,
                                    **kwargs)
         # True two-stage problem with proxy solution
        print("\n*** Solve Expectation of Proxy Distribution Problem (EPD) ***")
        exp_proxy_dist_sol = self.solve(scenario_tree_true, 
                                         decision_process=DecisionProcess(self._map_dvar_to_index, 
                                                                          {0: proxy_dist_sol.x0}),
                                         **kwargs)
        # True two-stage problem
        print("\n*** Solve True Stochastic Problem (SP) ***")
        true_stoch_sol = self.solve(scenario_tree_true, 
                                   decision_process=None,
                                   **kwargs)
        # value of proxy distribution
        if self._objective_sense == 'max':
            vpd = true_stoch_sol.objective_value - exp_proxy_dist_sol.objective_value
        else:
            vpd = exp_proxy_dist_sol.objective_value - true_stoch_sol.objective_value
        dict_vpd = {'PD': proxy_dist_sol,
                    'EPD': exp_proxy_dist_sol,
                    '2SP': true_stoch_sol,
                    'VPD': vpd,
                    'VPD(%)': 100 * vpd / (10**-10 + abs(true_stoch_sol.objective_value))}
        return dict_vpd
    
    # --- Precomputation ---
    @property
    def precomputed_decisions_name(self):
        return self._name_precomputed_decisions
    
    @property
    def precomputed_parameters_name(self):
        return self._name_precomputed_parameters
    
    def _initialize_precomputation(self):
        """Get the name of the precomputed parameters and variables. This sets the dictionary attributes 
        `_name_precomputed_decisions` and `_name_precomputed_parameters` at every stage."""
        for stage in range(self._n_stages):
            # decisions
            self._name_precomputed_decisions[stage] = set()
            if self.precompute_decision_variables(stage) is not None:
                for name, _ in self.precompute_decision_variables(stage):
                    self._name_precomputed_decisions[stage].add(name)
            # parameters
            self._name_precomputed_parameters[stage] = set()
            if self.precompute_parameters(stage) is not None:
                for name, _ in self.precompute_parameters(stage):
                    self._name_precomputed_parameters[stage].add(name)
            
    def precompute_decision_variables_at_node(self, node, model=None):
        if self._name_precomputed_decisions[node.level]: # implied booleanness tests if not empty
            if node.data.get('memory') is None:
                node.data['memory'] = {}
            self.set_path_info(node, model)
            for name, value_fct in self.precompute_decision_variables(node.level):
                self._memory_path[node.level][name] = value_fct()
        
    def precompute_parameters_at_node(self, node, verbose=0):
        if self._name_precomputed_parameters[node.level]: # implied booleanness tests if not empty
            if node.data.get('memory') is None:
                node.data['memory'] = {}
            self.set_path_info(node)
            for name, value_fct in self.precompute_parameters(node.level):
                if verbose:
                    print(f"\rPrecompute parameter '{name}' at node '{self._node_address}'..."+" "*20, end="")
                self._memory_path[node.level][name] = value_fct()
                                    
    # --- Sanity check ---        
    def check_sanity(self, scenario_tree, node_addresses=None):
        """ Check the validity of a scenario tree to solve the stochastic problem.
        
        Arguments:
        ----------
        scenario_tree: instance of ScenarioTree
            The scenario tree from which the information (scenario and decision) is taken.
            
        node_addresses: list of tuples (optional) (default: None)
            Each tuple is the address of a node for which sanity is checked.
            If None, sanity is checked at all nodes.
        """
        for node in scenario_tree.nodes:
            if node_addresses is not None and node.address not in node_addresses:
                continue
            self.set_path_info(node)
            self.sanity_check(node.level)
            
    def _check_scenario_trees(self, scenario_trees):
        """ Check the suitability between the problem and the scenario tree.
        
        Specifically, raise an Error if one of the following is True:
            (1) The scenario trees do not all have the same topology.
            (2) The number of random parameters as defined in the problem is larger than the one available 
                in the scenario tree.
        """
        assert len(set([tree.topology for tree in scenario_trees])) == 1, \
            f"Scenario trees should have the same topology to be solve together."
        for tree_index, scenario_tree in enumerate(scenario_trees, 1):
            for stage in range(1, self._n_stages):
                for var_name in self.map_rvar_name_to_nb[stage].keys():
                    n_var1 = self.map_rvar_name_to_nb[stage][var_name]
                    n_var2 = scenario_tree.map_stage_to_rvar_nb.get(stage, dict()).get(var_name, 0)
                    assert n_var1 <= n_var2, (f"Random parameter '{var_name}' has {n_var1} variables in the problem, "
                                              f"but only {n_var2} in the scenario tree #{tree_index}.")
                
    def _check_filenames(self, filename, which_filename, n_scenario_trees):
        if filename is None:
            return [None] * n_scenario_trees
        if isinstance(filename, str):
            assert n_scenario_trees == 1, \
                f"{which_filename} must be a list of strings if there are more than one scenario tree."
            return [filename]
        elif isinstance(filename, list):
            assert len(filename) == n_scenario_trees, (f"{which_filename} must contain as many strings as there are "
                  f"scenario trees, i.e., {n_scenario_trees}, not {len(filename)}.")
            return filename
        else:
            raise TypeError("{which_filename} should be a string or a list of strings.")
                
    # --- Representation ---
    def __repr__(self):
        dict_var_type = {'B': 'b', 'I': 'i', 'C': 'c'}
        str_ = f"{self._name}: ({self._n_stages} stages)\n"
        # dvar
        dvar_nb_list = [sum(self.map_dvar_name_to_nb[t].values()) for t in range(self._n_stages)]
        str_ += f"  - decision variables: {dvar_nb_list}\n"
        for stage in range(self._n_stages):
            str_ += f"    - stage {stage}: \n"
            for var, dict_ in self._map_dvar_to_index[stage].items():
                var_type = dict_var_type[self._map_stage_to_var_type[stage][var][0]]
                str_ += f"      - {var}: {len(dict_.values())} ({var_type})\n"
        # rvar
        rvar_nb_list = [sum(self.map_rvar_name_to_nb[t].values()) for t in range(self._n_stages)]
        str_ += f"  - random variables: {rvar_nb_list}\n"
        for stage in range(1, self._n_stages):
            str_ += f"    - stage {stage}: \n"
            for var, dict_ in self._map_rvar_to_index[stage].items():
                str_ += f"      - {var}: {len(dict_.values())}\n"
        return str_
    
    # ---  Copy, save and load ---
    def __copy__(self):
        cls = self.__class__
        new = cls.__new__(cls)
        new.__dict__.update(self.__dict__)
        return new
    
    def to_file(self, path, extension):
        import pickle
        with open(path, "wb") as f:
            pickle.dump(self, f)
   

def compare_scenario_trees(optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]], 
                           evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis], 
                           scenario_tree_opt1: ScenarioTree,
                           scenario_tree_opt2: ScenarioTree,
                           scenario_tree_eval: ScenarioTree) -> Dict[str, Union[float, StochasticSolutionBasis]]:
    """
    Optimize the same problem using two scenario trees and evaluate the two solutions in a third testing scenario tree.
       
    Arguments:
    ----------
    optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]]
        The optimizer that takes a scenario tree and returns the solution and its corresponding decision process.
        
    evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis]
        The evaluator that takes a scenario tree and a decision process and returns the solution.
        
    scenario_tree_opt1: ScenarioTree
        The scenario tree of the 1st uncertainty representation.
        
    scenario_tree_opt2: ScenarioTree
         The scenario tree of the 2nd uncertainty representation.
           
    scenario_tree_eval: ScenarioTree
        The scenario tree of the true uncertainty representation used to evaluate and compare the solutions.
               
    Returns:
    --------
    dict with keys and values:
         'SOL#1': instance of StochasticSolutionBasis (the solution with the 1st uncertainty)
         'SOL#2': instance of StochasticSolutionBasis (the solution with the 2nd uncertainty) 
         'E[SOL#1]':  instance of StochasticSolutionBasis (the evaluation of the solution of the 1st uncertainty)
         'E[SOL#2]':  instance of StochasticSolutionBasis (the evaluation of the solution of the 2nd uncertainty)
         'delta[#2-#1]': float 
         'delta[(#2-#1)/#1](%)': float 
    """
    # Problem scenario tree #1
    print("*** Solve Optimization Problem #1 (SOL#1) ***")
    sol1, dp1 = optimize_fct(scenario_tree_opt1)
    
    # Problem scenario tree #2
    print("\n*** Solve Optimization Problem #1 (SOL#2) ***")
    sol2, dp2 = optimize_fct(scenario_tree_opt2)
    
     # Evaluate solution of problem scenario tree #1
    print("\n*** Solve Expectation of Solution #1 (E[SOL#1]) ***")
    exp_sol1 = evaluate_fct(scenario_tree_eval, dp1)
    
     # Evaluate solution of problem scenario tree #2
    print("\n*** Solve Expectation of Solution #2 (E[SOL#2]) ***")
    exp_sol2 = evaluate_fct(scenario_tree_eval, dp2)
    
    # value of proxy distribution
    delta = exp_sol2.objective_value - exp_sol1.objective_value
    return {'SOL#1': sol1,
            'SOL#2': sol2,
            'E[SOL#1]': exp_sol1,
            'E[SOL#2]': exp_sol2,
            'delta[#2-#1]': delta,
            'delta[(#2-#1)/#1](%)': 100 * delta / (10**-10 + abs(exp_sol1.objective_value))}
    
    
def compute_value_right_dist(optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]], 
                            evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis], 
                            scenario_tree_true: ScenarioTree,
                            scenario_tree_proxy: ScenarioTree, 
                            objective_sense: str) -> Dict[str, Union[float, StochasticSolutionBasis]]:
    """ 
    Compute the Value of the Right Distribution (VRD).
    
    Note: VRD is defined as:
        VRD = v(EPD) - v(SP) (for mininization) or v(SP) - v(EPD) (for maximization)
    where:
        v(SP) is the optimal-value of the Two-Stage Problem (SP)
        v(EPD) is the optimal-value of the Expectation of Proxy Distribution Problem (EPD)
    The EPD problem evaluates the value of the Proxy Distribution (PD) solution in the true stochastic problem:
        v(EPD) = v(SP s.t. x0 = x0(PD))
    where x0(PD) is the stage-0 optimal solution obtained using the proxy scenarios.
       
    Arguments:
    ----------
    optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]]
        The optimizer that takes a scenario tree and returns the solution and its corresponding decision process.
        
    evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis]
        The evaluator that takes a scenario tree and a decision process and returns the solution.
    
    scenario_tree_true: ScenarioTree
        The scenario tree of the true distribution (SP).
        
    scenario_tree_proxy: ScenarioTree
        The scenario tree of the proxy distribution (PD).
                  
    objective_sense: {'max', 'min'}
        
    Returns:
    --------
    Dict[str, Union[float, StochasticSolutionBasis]]: mapping:
         'PD': instance of StochasticSolutionBasis (the PD solution instance)
         'EPD': instance of StochasticSolutionBasis (the EPD solution instance)
         'SP':  instance of StochasticSolutionBasis (the SP solution instance)
         'VRD': float (the Value of Right Distribution)
         'VRD%' float (the Value of Right Distribution in percentage of v(2SP))
    """
    # Problem with proxy scenarios
    print("*** Solve Proxy Distribution Problem (PD) ***")
    proxy_dist_sol, dec_pro = optimize_fct(scenario_tree_proxy)
    
     # True two-stage problem with proxy solution
    print("\n*** Solve Expectation of Proxy Distribution Problem (EPD) ***")
    exp_proxy_dist_sol = evaluate_fct(scenario_tree_true, dec_pro)
    
    # True two-stage problem
    print("\n*** Solve True Stochastic Problem (SP) ***")
    true_stoch_sol, dec_pro = optimize_fct(scenario_tree_true)
    
    # value of proxy distribution
    if objective_sense == 'max':
        vrd = true_stoch_sol.objective_value - exp_proxy_dist_sol.objective_value
    else:
        vrd = exp_proxy_dist_sol.objective_value - true_stoch_sol.objective_value
    return {'PD': proxy_dist_sol,
            'EPD': exp_proxy_dist_sol,
            'SP': true_stoch_sol,
            'VRD': vrd,
            'VRD(%)': 100 * vrd / (10**-10 + abs(true_stoch_sol.objective_value))}
    
    
def compute_wait_and_see(optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]],
                         scenario_tree: ScenarioTree) -> List[Tuple[StochasticSolutionBasis, DecisionProcess]]:
    """ 
    Solve the Wait-and-See (WS) problems. Each WS problem is formulated by taking a single scenario in 
    the scenario tree. 
    
    Arguments:
    ----------
    optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]]
        The optimizer that takes a scenario tree and returns the solution and its corresponding decision process.
        
    scenario_tree: ScenarioTree
        The scenario tree of representing the stochastic problem uncertainty.

    Returns:
    --------
    List[Tuple[StochasticSolutionBasis, DecisionProcess]]
    """
    return [optimize_fct(scenario_tree) for scenario_tree in decompose(scenario_tree)]
    

def compute_opportunity_cost(optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]],
                             evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis],
                             scenario_tree: ScenarioTree) -> np.ndarray:            
    """
    Compute the square-matrix of opportunity cost. The computation consists in two steps:
        (1) solving the two-stage problem for each scenario independently;
        (2) solving the 2nd-stage problem for each scenario and each 1st-stage solution computed at step (1).
    
    Arguments:
    ----------
    optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess]]
        The optimizer that takes a scenario tree and returns the solution and its corresponding decision process.
        
    evaluate_fct: Callable[[ScenarioTree, DecisionProcess], StochasticSolutionBasis]
        The evaluator that takes a scenario tree and a decision process and returns the solution.
        
    scenario_tree: ScenarioTree
        The scenario tree of representing the stochastic problem uncertainty.

    Returns:
    --------
    2d-array of shape (n_leaves, n_leaves) where n_leaves is the number of leaves in the scenario tree. 
        The element [i, j] is the objective value of the i-th wait-and-see solution when evaluated in the j-th 
        scenario. 
        Formally, it is: F(x_i^*, xi_j) 
        where x_i^* is the solution of the i-th WS problem and xi_j is the j-th scenario of the scenario tree.
    """
    # (1) Solve wait and see problems
    ws_solutions = []
    for k, scenario_tree in enumerate(decompose(scenario_tree)):
        with TimeIt(f"Solve Wait-and-See Problems #{k}...\n", lambda t: f"Finished {t:.3f} sec\n"):
            ws_solutions.append(optimize_fct(scenario_tree))
     
    # (2) Evaluate wait and see solutions
    n_leaves = scenario_tree.width[-1]
    matrix = np.zeros((n_leaves, n_leaves))
    for k, (ws_sol, ws_dp) in enumerate(ws_solutions):
        with TimeIt(f"Evaluate Wait-and-See Solution #{k}...", lambda t: f"Finished {t:.3f} sec\n"):
            eval_sol = evaluate_fct(scenario_tree, ws_dp)
            for j, leaf in enumerate(eval_sol.scenario_tree.leaves):
                matrix[k, j] = leaf.data['v']
    return matrix





# -*- coding: utf-8 -*-

from typing import Callable, Tuple, Optional, Dict, Union

import numpy as np
from scipy.stats import norm
 
from stochoptim.scengen.tree_structure import Node

ScenarioType = Dict[str, np.ndarray]
ScenarioPathType = Dict[int, ScenarioType]

class ScenarioProcess:
    """The process that fills a tree structure with scenarios.
    
    Arguments:
    ----------
    scenario_fct: Callable[[int, 1d-array, Dict[int, Dict[str, 1d-array]]], Dict[str, 1d-array]]
        Function that takes as argument the stage, the discretization point (called epsilon by convention), 
        and the scenario path up to (but not including) the stage, and returns the corresponding scenario.
        (See the signature of the `get_scenario` method.)
        
    epsilons_fct: Callable[[int, int], 2d-array] or None (default: None)
        Function that takes as argument the stage and the number of sample points, and returns a sample of 
        discretization points.
        (See the signature of the `get_epsilon_sample` method.)
        
    weights_fct: Callable[[int, int], 1d-array] or None (default: None)
        Function that takes as argument the stage and the number of sample points, and returns a sample of 
        discretization weights (probabilities if they sum to one).
        (See the signature of the `get_epsilon_sample` method.)
        
    name: str or None (default: None)
        Name of the scenario process
        
    checker: bool (default: False)
        If True, the type and size of the scenarios and the discretization points and weights are checked.
        
    stochastic_problem: instance of a subclass of StochasticProblemBasis or None (default: None)
        If provided (and if `checker` is True) the scenarios are checked to be in accordance with what is specified
        in the problem in terms of the names of the random variables and their dimensions.
    """
    
    def __init__(self, 
                 scenario_fct: Callable[[int, np.ndarray, ScenarioPathType], ScenarioType], 
                 epsilons_fct: Optional[Callable[[int, int], np.ndarray]] = None, 
                 weights_fct: Optional[Callable[[int, int], np.ndarray]] = None,
                 name: Optional[str] = None,
                 checker: bool = True,
                 stochastic_problem: Optional['StochasticProblemBasis'] = None):  
        """

        """
        self._scenario_fct = scenario_fct
        self._epsilons_fct = epsilons_fct
        self._weights_fct = weights_fct
        self._name = name
        self._checker = checker
        self._map_rvar_name_to_nb = stochastic_problem.map_rvar_name_to_nb if stochastic_problem is not None else None
    
    def get_epsilon_sample(self, 
                           n_samples: int, 
                           stage: int) -> Tuple[np.ndarray, Optional[np.ndarray]]:   
        """Return the discretization points and weights of the stagewise independent random vector epsilon.
        
        Arguments:
        ----------
        n_samples: int >= 1
            Number of points (and weights) to be generated.
            
        stage: int >= 0
            Stage of the random variable epsilon to be discretized.
            
        Returns:
        --------
        Tuple (1d-array, 2d-array) where:
            
        1d-array: the discretization weights given as an array of shape (n_samples,).
                
        2d-array or None: the discretization points given as an array of shape (n_samples, dim_epsilon), for an 
        arbitrary value of dim_epsilon.
        """
        weights = self._weights_fct(n_samples, stage) if self._weights_fct is not None else np.ones(n_samples) / n_samples
        epsilons = self._epsilons_fct(n_samples, stage) if self._epsilons_fct is not None else None
        if self._checker:
            assert isinstance(weights, np.ndarray), \
                f"The discretization weights must be given as a 1d numpy array, not as type {type(weights)}"
            assert len(weights.shape) == 1, \
                f"The numpy array of weights must be a 1-dimensional, not of dimension {len(weights.shape)}"
            if epsilons is not None:
                assert isinstance(epsilons, np.ndarray), \
                    f"The discretization points (epsilons) must be given as a 2d numpy array, not as type {type(epsilons)}"
                assert len(epsilons.shape) == 2, \
                    f"The numpy array of discretization points (epsilons) must be 2-dimensional, not {len(epsilons.shape)}-dimensional"
                assert weights.shape[0] == epsilons.shape[0], ("Mismatch between the number of discretization weights "
                f"({weights.shape[0]}) and points ({epsilons.shape[0]}).")
        return weights, epsilons
    
    def get_scenario(self, 
                     stage: int, 
                     epsilon: Optional[np.ndarray], 
                     scenario_path: Optional[ScenarioPathType]) -> ScenarioType:
        r"""Return the scenario value given as a function of the epsilon sample point (the current stage contribution) 
        and the scenario path (the contribution of the history up to the current stage). 
        
        Mathematically it is the implementation of the recursion: 
            \xi_{0} = initial value,
            \xi_{stage} = g_t(\epsilon_{stage}, \xi_{..stage-1}), for stage=1,...
        
        Arguments:
        ----------           
        stage: int >= 0
            Current stage of the random parameters to be discretized.
            
        epsilon: 1d-array or None
            Discretization point of epsilon_{stage}, of shape (dim_epsilon,).
            These discretization points are the ones returned by the `get_epsilon_sample` method.
            
        scenario_path: Dict[int, Dict[str, 1d-array]] or None
            History of the process up to (but not including) the current stage.
            
        Returns:
        --------           
        Dict[str, 1d-array]: the discretization point of \xi_{stage}, of shape (dimension_of_\xi_{stage},)
        """
        scenario = self._scenario_fct(stage, epsilon, scenario_path)
        if self._checker and stage >= 1: # we don't check the scenario at stage 0 (at the root)
            assert isinstance(scenario, dict), \
                ("A scenario must be a dictionary mapping the variable names (str) to the variable arrays (1d-array)\n"
                 f"Here scenario was of type {type(scenario)}: {scenario}")
            if self._map_rvar_name_to_nb is not None:
                assert set(scenario.keys()) == set(self._map_rvar_name_to_nb[stage].keys()), \
                    (f"Mismatch between the variable names given in the problem at stage {stage}: "
                     f"{set(self._map_rvar_name_to_nb[stage].keys())}, "
                     f"and those found in the scenario process: {set(scenario.keys())}")
            for var_name, var_array in scenario.items():
                assert isinstance(var_array, np.ndarray), \
                    f"The variables must be given as a 1d numpy array, not as type {type(var_array)}"
                assert len(var_array.shape) == 1, \
                    f"The numpy array must be a 1-dimensional, not of dimension {len(var_array.shape)}"
                if self._map_rvar_name_to_nb is not None:
                    assert var_array.shape[0] == self._map_rvar_name_to_nb[stage][var_name], \
                    (f"Mismatch between the number of components of variable '{var_name}' given in the problem at "
                     f"stage {stage} ({self._map_rvar_name_to_nb[stage][var_name]}) "
                     f"and the number found in the scenario process ({var_array.shape[0]})")
        return scenario
    
    def get_children_sample(self, 
                            node: Node) -> Tuple[np.ndarray, np.ndarray]:
        """Same as .get_epsilon_sample() but callable on a node."""
        assert not node.is_leaf, "Cannot get the children epsilon sample of a leaf"
        return self.get_epsilon_sample(len(node.children), node.level + 1)
        
    def get_node_scenario(self, 
                          node: Node, 
                          path: bool) -> Union[ScenarioType, ScenarioPathType]:
        """Returns the scenario at a node or the scenario path leading to that node.
        
        Arguments:
        ----------
        node: Node
        
        path: bool
            If True, then the scenario path from the root to 'node' is returned.
            Otherwise only the scenario at 'node' is returned.
            
        Returns:
        -------
        Dict[str, 1d-array] (if path is False)
        Dict[int, Dict[str, 1d-array]] (if path is True)
            The scenario at the node or the scenario path leading to the node.
        """
        if path:
            return {stage: self.get_node_scenario(n, path=False) 
                    for stage, n in enumerate(node.branch)}
        else:
            return self.get_scenario(node.level, 
                                     node.data.get("eps"), 
                                     Node.get_data_path(node.parent, 'scenario') 
                                     if not node.is_root else None)


def mc_normal_sample(n_samples, mu=0, sigma=1):
    return np.ones(n_samples) / n_samples, np.random.normal(loc=mu, scale=sigma, size=(n_samples, 1))


def qmc_normal_sample(n_samples, mu=0, sigma=1, u=0.5):
    return np.ones(n_samples) / n_samples, \
            mu + sigma * norm.ppf((np.linspace(0, 1-1/n_samples, n_samples) + u) % 1).reshape(-1, 1)    
        


        



# -*- coding: utf-8 -*-

import numpy as np
from typing import Dict, List, Optional

from stochoptim.scengen.decision_process import DecisionProcess
from stochoptim.scengen.scenario_tree import ScenarioTree

class StochasticSolutionBasis:
        
    def __init__(self, 
                 stochastic_problem,
                 scenario_tree):
        
        self._stochastic_problem = stochastic_problem
        self._scenario_tree = scenario_tree
            
    # --- Properties ---
    @property
    def stochastic_problem(self):
        """The stochastic problem"""
        return self._stochastic_problem
        
    @property
    def scenario_tree(self):
        """The scenario tree"""
        return self._scenario_tree
    
    @property
    def x0(self):
        """The stage-0 decisions"""
        return self._scenario_tree.data.get("decision")
        
    @property
    def status(self):
        return self._scenario_tree.data.get('status')
    
    @property
    def objective_bound(self):
        return self._scenario_tree.data.get('bound', np.nan)

    @property
    def objective_value(self):
        return self._scenario_tree.data.get('v', np.nan)

    @property
    def solve_time(self):
        return self._scenario_tree.data.get('time', np.nan)
    
    @property    
    def objective_value_at_leaves(self):
        return np.array([leaf.data.get('v', np.nan) for leaf in self._scenario_tree.leaves])
    
    @property
    def gap(self):
        return abs(self.objective_bound - self.objective_value)
    
    @property
    def gap_percent(self):
        return 100 * self.gap / abs(10**(-10) + self.objective_value)
    
    @property
    def n_scenarios(self):
        return sum(1 for _ in self._scenario_tree.leaves)
    
    def warmstart(self, excluding_var_names: Optional[Dict[int, List[str]]] = None):
        """Build a warmstart dictionary from the scenario-tree solution.
        This dictionary can be given as argument of StochasticProblemBasis.solve() under the 
        keyword 'warmstart'.
        
        Returns:
        --------
        dictionary mapping variable names (str) to values (int or float).
        """
        if excluding_var_names is None:
            excluding_var_names = dict()
        warmstart_dict = {}
        for node in self._scenario_tree.nodes:
            for var_name, map_subcript_to_index in self._stochastic_problem.map_dvar_to_index[node.level].items():
                if var_name in excluding_var_names.get(node.level, []):
                    continue
                for subscript, index in map_subcript_to_index.items():
                    if isinstance(subscript, tuple):
                        subscript_str = '_'.join([str(s) for s in subscript])
                    else:
                        subscript_str = str(subscript)
                    key = f'{var_name}_{node.address}_{subscript_str}'
                    var_type = self._stochastic_problem.map_stage_to_var_type[node.level][var_name][0]
                    if var_type in ['B', 'I']:
                        warmstart_dict[key] = int(np.round(node.data['decision'][var_name][index]))
                    else:
                        warmstart_dict[key] = float(node.data['decision'][var_name][index])
        return warmstart_dict
    
    # --- To display solution ---
    def set_path_info(self, scen_index: Optional[int] = None):
        if scen_index is None: # root
            self._stochastic_problem.set_path_info(self._scenario_tree)
        else:    
            self._check_scenario_index_validity(scen_index)
            leaf = next(leaf for k, leaf in enumerate(self._scenario_tree.leaves) if k == scen_index)
            self._stochastic_problem.set_path_info(leaf)

    def decision_path(self, scen_index: int) -> Dict[int, Dict[str, np.ndarray]]:
        """Return a path of decisions from the root to a leaf"""
        self._check_scenario_index_validity(scen_index)
        self.set_path_info(scen_index)
        return self._stochastic_problem._decision_path
       # leaf = next(leaf for k, leaf in enumerate(self._scenario_tree.leaves) if k == scen_index)
       # return self._scenario_tree.get_data_path(leaf, 'decision')
  
    def decision_process_fixed(self, 
                               var_names: Dict[int, List[str]],
                               scen_index: Optional[int] = None) -> DecisionProcess:
        """Return a decision process made of a selection of decision variables at each stage. 
        (The decision process is said to be fixed because it does not depend on the scenario path.)
        
        var_names:
            Dictionary mapping a stage to the list of variable names to be included in the decision process
            at that stage.
        """
        assert not (scen_index is None and list(var_names.keys()) != [0]), \
            "Only variables at stage 0 can be considered if `scen_index` is None."
        # generate empty decision process
        decision_process = DecisionProcess(self._stochastic_problem._map_dvar_to_index)
        # get decision path on the required scen_index
        decision_path = self.decision_path(scen_index) if scen_index is not None else {0: self.x0}
        # add one by one the required variables to the process 
        for stage in var_names.keys():
            for var_name in var_names[stage]:
                decision_process.update_decision_array(decision_path[stage][var_name], stage, var_name)
        return decision_process
    
    def generate_empty_decisions(self):
        for stage, map_name_to_nb in self.stochastic_problem.map_dvar_name_to_nb.items():
            for node in self.scenario_tree.nodes_at_level(stage):
                node.data['decision'] = {}
                for name, nb in map_name_to_nb.items():
                    node.data['decision'][name] = np.empty(nb)          
            
    # --- Sanity ---
    def _check_scenario_index_validity(self, scen_index):
        assert scen_index in range(self.n_scenarios), \
                (f"Scenario index should be between 0 and {self.n_scenarios-1}, not {scen_index}.")
                
    def _check_x0_validity(self, x0):
        assert len(x0) == sum(self.stochastic_problem.map_dvar_name_to_nb[0].values()), \
            (f"Sanity check failed: mistmatch between the dimension of x0 ({len(x0)}) "
                "and the theoretical number of stage-0 decisions of self.stochastic_problem "
                f"({sum(self.stochastic_problem.map_dvar_name_to_nb[0].values())}")
        
    def _check_sanity(self):
        self._check_x0_validity(self.x0)
       
    # --- Representations ---
    def __repr__(self):
        return (f"obj={self.objective_value:,.3f}, "
                f"gap={self.gap_percent:.3f}%, "
                f"time={self.solve_time:.1f} sec")
        
    # ---- Load, save, copy ---                   
    @classmethod
    def from_file(cls, 
                  problem_class_or_module,
                  path_tree, 
                  path_problem, 
                  extension_tree='pickle', 
                  extension_prob='txt'):
        """ Load network design solution from two files: one for the problem and one for the scenario tree.
        Arguments:
        ----------
        problem_class_or_module: class or module containing the stochastic problem.
            This must have a method/function `from_file`.
            
        path_tree: str
        
        extension_tree: {'pickle', 'txt'}
        
        path_problem: str
        
        extension_prob: {'pickle', 'txt'}
        Returns:
        --------
        instance of StochasticSolutionBasis or a subclass
        """            
        assert hasattr(problem_class_or_module, 'from_file'), \
            "The problem class or module given as input `problem_class_or_module` must have a method `from_file()`."
        problem = problem_class_or_module.from_file(path_problem, extension_prob)
        scenario_tree = ScenarioTree.from_file(path_tree, extension_tree)
        return cls(problem, scenario_tree)
    
    def to_file(self, 
                path_tree,
                extension_tree, 
                path_problem,
                extension_prob, 
                with_keys=None, 
                without_keys=None):
        """ Save the scenario tree (containing the optimal decisions) and the stochastic problem in two
        separated files.
        
        Arguments:
        ----------
        path_tree: string
            Path to the file containing the scenario tree information.
            
        path_prob: string
            Path to the file containing the stochastic problem information.
            
        extension_tree: {'txt', 'pickle'}
            Format in which the scenario tree is solved.
            
        extension_prob: {'txt', 'pickle'}
            Format in which the stochastic problem is solved.
        
        with_keys: list of strings or None
            If not None, it is the data keys to be saved in the scenario tree. 
            If None, then all keys in the list will be saved except those in 'without_keys'.
            
        without_keys: list of strings or None
            This argument is used only if 'with_keys' is None. 
            In that case it contains the data keys not to be saved in the scenario tree.
        """
        self._scenario_tree.to_file(path_tree, extension_tree, with_keys, without_keys)
        self._stochastic_problem.to_file(path_problem, extension_prob)


# -*- coding: utf-8 -*-

from typing import Dict, Union, Tuple, Callable, List, Optional

import numpy as np

from stochoptim.scengen.scenario_tree import ScenarioTree
from stochoptim.scengen.tree_structure import Node

# shortcuts for important types 
Num = Union[int, float]
SubscriptType = Union[int, str, Tuple[Union[int, str], ...]] # variable subscript
FixedDecisionType = Optional[Num]
ScenarioPathType = Dict[int, Dict[str, np.ndarray]]
CallableDecisionType = Callable[[ScenarioPathType], FixedDecisionType]
DecisionType = Union[FixedDecisionType, CallableDecisionType]


class DecisionProcess:
    """The process that maps scenario paths to decisions"""
    
    def __init__(self, 
                 map_dvar_to_index: Dict[int, Dict[str, Dict[SubscriptType, int]]], 
                 input_decision_dict: Optional[Dict[int, Dict[str, List[DecisionType]]]] = None,
                 checker=False):
        """
        Arguments:
        ----------
        map_dvar_to_index: Dict[int, Dict[str, Dict[SubscriptType, int]]]
            The mapping from the stages to the decision variables that defines the stochastic problem.
            The mapping is: stage (int) -> variable name (str) -> variable subscript (SubscriptType) 
                                                                -> index in the array (int)
            
        input_decision_dict: Dict[int, Dict[str, List[DecisionType]]] or None (default: None)
            The decisions that are fixed beforehand. (More fixed decisions can be added later on via the 
            `update_decision_array` and `update_decision_value` methods)
            
        checker: bool (default: True)
            If True, check that the decision variables have the appropriate size (as specified
            in the stochastic problem) when being updated.
        """
        self._map_dvar_to_index = map_dvar_to_index
        self._map_stage_to_dvar_nb: Dict[int, Dict[str, int]] = {}
        self._decision_dict: Dict[int, Dict[str, List[DecisionType]]] = {}
        self._checker = checker
        self._initialize_decision_dict()
        
        if input_decision_dict is not None:
            self.update_decision_dict(input_decision_dict)                
                    
    # --- Properties ---
    @property
    def n_stages(self):
        return len(self._map_stage_to_dvar_nb.keys())
    
    @property
    def get_dict(self):
        return self._decision_dict
    
    @property
    def map_stage_to_dvar_nb(self):
        return self._map_stage_to_dvar_nb
    
    @property
    def map_dvar_to_index(self):
        return self._map_dvar_to_index
    
    # --- Initialization ---
    def _initialize_decision_dict(self):
        for stage in self._map_dvar_to_index.keys():
            self._decision_dict[stage] = {}
            self._map_stage_to_dvar_nb[stage] = {}
            for var_name, map_subscript_to_index in self._map_dvar_to_index[stage].items():
                self._map_stage_to_dvar_nb[stage][var_name] = len(map_subscript_to_index.values())
                self._decision_dict[stage][var_name] = np.array([None] * self._map_stage_to_dvar_nb[stage][var_name])
           
    # --- Update decisions ---         
    def update_decision_dict(self, 
                             new_decision_dict: Dict[int, Dict[str, List[DecisionType]]]):
        for stage in self._map_dvar_to_index.keys():
            if new_decision_dict.get(stage) is None:
                continue
            for var_name in self._map_dvar_to_index[stage].keys():
                array = new_decision_dict[stage].get(var_name)
                if array is not None:
                    self.update_decision_array(array, stage, var_name)   
                        
    def update_decision_array(self, 
                              new_decision_array: List[DecisionType], 
                              stage: int, 
                              var_name: str):
        """Update a whole array of decisions"""
        new_decision_array = np.array(new_decision_array)
        assert len(new_decision_array.shape) == 1, "The new decision array should be 1-dimensional"
        if self._checker:
            assert new_decision_array.shape[0] == self._map_stage_to_dvar_nb[stage][var_name], \
                (f"The new decision array should have the same size as the previous one, namely "
                 f"{(self._map_stage_to_dvar_nb[stage][var_name],)}, not {new_decision_array.shape}.")
        self._check_var_name(stage, var_name)
        self._decision_dict[stage][var_name] = new_decision_array
        
    def update_decision_value(self, 
                                new_decision: DecisionType, 
                                stage: int, 
                                var_name: str, 
                                var_subscript: SubscriptType):
        """Update a single decision"""
        assert new_decision is None or isinstance(new_decision, (int, float)) or callable(new_decision), \
            f"The new decision should an int, float, None or a callable, not {type(new_decision)}"
        self._check_var_name(stage, var_name)
        index = self._map_dvar_to_index[stage][var_name][var_subscript]
        self._decision_dict[stage][var_name][index] = new_decision
            
    def _check_var_name(self, stage, var_name):
        assert var_name in self._map_dvar_to_index[stage].keys(), (f"Variables {var_name} does not exist at stage "
            f"{stage}, should be one of {list(self._map_dvar_to_index[stage].keys())}")
        
    # --- Get decisons ---               
    def _callable_mask(self, stage, var_name):
        """Returns a mask on the variables with True if it is callable"""
        assert isinstance(self._decision_dict[stage][var_name], np.ndarray), \
            f"Decisions must be of type numpy.ndarray, not {type(self._decision_dict[stage][var_name])}"
        # if type is not object then all decisions are int/float
        if self._decision_dict[stage][var_name].dtype != np.dtype('O'): 
            return np.zeros((self._map_stage_to_dvar_nb[stage][var_name]), dtype='bool')
         # if all decisions are None then none are callable
        if (self._decision_dict[stage][var_name] == None).all():
            return np.zeros((self._map_stage_to_dvar_nb[stage][var_name]), dtype='bool')
        # if some decision are not None, then we check one by one if they are callable
        else: 
            return np.vectorize(callable)(self._decision_dict[stage][var_name])
    
    def get_decision_value(self, 
                            stage: int, 
                            var_name: str, 
                            var_subscript: SubscriptType,
                            scenario_path: Optional[ScenarioPathType] = None) -> FixedDecisionType:
        """Return a single decision"""
        index = self._map_dvar_to_index[stage][var_name][var_subscript]
        if callable(self._decision_dict[stage][var_name][index]):
            return self._decision_dict[stage][var_name][index](scenario_path)
        else:
            return self._decision_dict[stage][var_name][index]
        
    def get_decision_array(self,
                            stage: int,
                            var_name: str,
                            scenario_path: Optional[ScenarioPathType] = None) -> List[FixedDecisionType]:
        """Return a whole array of decisions"""
        callable_mask = self._callable_mask(stage, var_name)
        if not callable_mask.any(): # if no variable is callable
            return self._decision_dict[stage][var_name]
        else:
            decision_array = self._decision_dict[stage][var_name]
            for index in callable_mask.nonzero()[0]:
                decision_array[index] = self._decision_dict[stage][var_name][index](scenario_path)
            return decision_array
        
    def get_decision_dict(self, 
                           stage: int, 
                           scenario_path: Optional[ScenarioPathType] = None) -> Dict[str, List[FixedDecisionType]]:
        """Return a whole dictionary of decisions at one stage"""
        return {var_name: self.get_decision_array(stage, var_name, scenario_path) 
                             for var_name in self._decision_dict[stage].keys()}
                            
    def __call__(self, 
                 node: Node, 
                 var_name: Optional[str] = None, 
                 var_subscript: Optional[SubscriptType] = None) -> Union[FixedDecisionType, 
                                                                         List[FixedDecisionType], 
                                                                         Dict[str, List[FixedDecisionType]]]:
        if var_name is None and var_subscript is None:
            return self.get_decision_dict(node.level, ScenarioTree.get_scenario_path(node))
        
        if var_name is not None and var_subscript is None:
            return self.get_decision_array(node.level, var_name, ScenarioTree.get_scenario_path(node))
        
        if var_name is not None and var_subscript is not None:
            return self.get_decision_value(node.level, var_name, var_subscript, ScenarioTree.get_scenario_path(node))
        
    # --- Representation ---
    def __repr__(self):
        str_ = f"Decision process: ({self.n_stages} stages)\n"
        # dvar
        dvar_nb_list = [sum(self._map_stage_to_dvar_nb[t].values()) for t in range(self.n_stages)]
        str_ += f"  - decision variables: {dvar_nb_list}\n"
        for stage in range(self.n_stages):
            str_ += f"    - stage {stage}: \n"
            for var_name, dict_ in self._map_dvar_to_index[stage].items():
                n_fixed = (self._decision_dict[stage][var_name] != None).sum()
                n_callable = self._callable_mask(stage, var_name).sum()
                if n_fixed > 0 or n_callable > 0:
                    str_ += f"      - {var_name}: {len(dict_.values())} (fixed: {n_fixed}, callable: {n_callable})\n"
                else:
                    str_ += f"      - {var_name}: {len(dict_.values())}\n"
        return str_

    @classmethod
    def from_problem(cls, 
                     stochastic_problem, 
                     input_decision_dict: Optional[Dict[int, Dict[str, List[DecisionType]]]] = None):
        assert hasattr(stochastic_problem, 'map_dvar_to_index'), ("The stochastic problem "
        "must have an attribute `map_dvar_to_index`")
        return cls(stochastic_problem.map_dvar_to_index, input_decision_dict)        

from_problem = DecisionProcess.from_problem


# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
from itertools import product
import numpy as np

from ..stochastic_solution_basis import StochasticSolutionBasis


class FacilityLocationSolution(StochasticSolutionBasis):
    
    def __init__(self, 
                 stochastic_problem,
                 scenario_tree):
        
        StochasticSolutionBasis.__init__(self, 
                                         stochastic_problem,
                                         scenario_tree)

    # --- facilities activity ---
    def is_active(self, facility=None):
        self.set_path_info()
        return self.stochastic_problem.x(facility)
        
    def n_active_facilities(self):
        return np.sum(self.is_active())
    
    # --- facility penalty ---
    def get_penalty_value(self, scen_index, facility=None):
        """Returns the penality receives by a facility in a given scenario"""
        self.set_path_info(scen_index)
        return self.stochastic_problem.z(facility)
    
    def is_penalized(self, scen_index=None, facility=None):
        """Returns True if the facility is penalized in at least one of the scenario, False otherwise"""
        if scen_index is not None:
            return self.get_penalty_value(scen_index, facility) > 0
        else:
            return (np.sum([self.get_penalty_value(scen_index, facility) 
                            for scen_index in range(self.n_scenarios)], axis=0) > 0).astype('bool')
        
    # --- Clients presence ---
    def is_present(self, scen_index, client=None):
        """Returns True if a client location is occupied in a given scenario"""
        self.set_path_info(scen_index)
        return self.stochastic_problem.h(client)
    
    # --- facility/Client assignement ---
    def is_assigned(self, scen_index, client=None, facility=None):
        self.set_path_info(scen_index)
        return self.stochastic_problem.y(client, facility)
    
    # --- Plots ---
    def _plot_facilities(self, scen_index, ax):
        pos_facilities = self.stochastic_problem.pos_facility()
        penalized_mask = self.is_penalized(scen_index)
        active_not_penalized_mask = np.multiply(self.is_active(), ~self.is_penalized(scen_index)).astype('bool')
        not_active_mask = ~(self.is_active().astype('bool'))
        # if facility location is active and penalized in at least one scenario
        if penalized_mask.sum() > 0:
            ax.scatter(*zip(*pos_facilities[penalized_mask]), marker="s", s=30, c="k")
        # if facility location is active but never penalized
        if active_not_penalized_mask.sum() > 0:
            ax.scatter(*zip(*pos_facilities[active_not_penalized_mask]), marker="s", s=30, c="C1")    
        # if facility location is not active
        if not_active_mask.sum() > 0:
            ax.scatter(*zip(*pos_facilities[not_active_mask]), marker="s", s=30, alpha=0.8, 
                       facecolors='none', edgecolors='C1')   
                
    def _plot_clients(self, scen_index, ax):
        is_present_pos = self.stochastic_problem.pos_client()[self.is_present(scen_index).astype('bool')]
        is_not_present_pos = self.stochastic_problem.pos_client()[~(self.is_present(scen_index).astype('bool'))]
        # client is present at location
        if is_present_pos.sum() > 0: 
            ax.scatter(*zip(*is_present_pos), marker="o", c="C0", s=20)
        # client not present at location
        if is_not_present_pos.sum() > 0:
            ax.scatter(*zip(*is_not_present_pos), marker="o", facecolors='none', edgecolors='C0', s=20)
             
    def _plot_clients_facilities_connection(self, scen_index, ax):
        for c, s in product(self.stochastic_problem.client_locations, self.stochastic_problem.facility_locations):
            if self.is_assigned(scen_index, c, s):
                ax.plot(*zip(self.stochastic_problem.pos_client(c), self.stochastic_problem.pos_facility(s)), 
                        color="C{}".format(s % 10))
                        
    def plot_network(self, scen_index=None, show_clients=True, show_facilities=True, show_title=True, 
                     figsize=(5,5), path=None, extension=".pdf", **kwargs):
        
        fig, ax = plt.subplots(figsize=figsize)
        # plot facilities locations
        if show_facilities:
            self._plot_facilities(scen_index, ax)
            
        if scen_index is None:
            if show_clients:
                ax.scatter(*zip(*self.stochastic_problem.pos_client()), marker="o", 
                           facecolors='none', edgecolors='C0', s=20)
        else:
            self._plot_clients(scen_index, ax)
            if show_facilities:
                # plot links between facilities and clients assigned to them
                self._plot_clients_facilities_connection(scen_index, ax)
                        
        if show_title:   
            if scen_index is None:
                obj_to_show = self.objective_value
                ax.set_title(f"$v^*: {obj_to_show:.2f}$")
            else:
                obj_to_show = self.objective_value_at_leaves[scen_index]
                ax.set_title(f"$v^*: {obj_to_show:.2f}$")         
             
        self._set_axes(ax, **kwargs)
        self._save_fig(path, extension)
        plt.show()  
            
    def _save_fig(self, path, extension):
        if path:
            plt.savefig(path + extension)
            print(f"Figure saved at {path + extension}")
                
    def _set_axes(self, ax, xlim=None, ylim=None, show_xaxis=False, show_yaxis=False, show_frame=True):
        # set x and y limits
        if xlim is not None:
            ax.set_xlim(*xlim)
        if ylim is not None:
            ax.set_ylim(*ylim)   
        # show frame and axes
        ax.get_xaxis().set_visible(show_xaxis)
        ax.get_yaxis().set_visible(show_yaxis)
        if not show_frame and not show_xaxis and not show_yaxis:
            ax.axis('off')
    
    @classmethod           
    def from_file(cls, 
                  path_tree, 
                  path_problem, 
                  extension_tree='pickle', 
                  extension_prob='txt'):
        from stochoptim.stochprob.facility_location.facility_location_problem import FacilityLocationProblem
        return StochasticSolutionBasis.from_file(FacilityLocationProblem, 
                                                 path_tree, path_problem, 
                                                 extension_tree, extension_prob)
        
from_file = FacilityLocationSolution.from_file


# -*- coding: utf-8 -*-

import sys
import os
import time

import numpy as np

numpy_integer_types  = [np.dtype(f'int{bits}') for bits in [8, 16, 32, 64]]
                         
numpy_float_types = [np.dtype(f'float{bits}') for bits in [16, 32, 64]] 

numpy_numerical_types  = numpy_integer_types + numpy_float_types
            
def progress_bar(value, endvalue, bar_length=20, endline=""):
    percent = float(value) / endvalue
    arrow = '-' * int(round(percent * bar_length)-1) + '>'
    spaces = ' ' * (bar_length - len(arrow))
    sys.stdout.write("\rPercent: [{0}] {1}%".format(arrow + spaces, int(round(percent * 100))))
    sys.stdout.flush()
    if value == endvalue:
        sys.stdout.write(endline)    
    
class TimeIt(object):
    
    base = [0] * 100
    
    def __init__(self, 
                 string_enter, 
                 string_exit_callable, 
                 level_enter=0, 
                 level_exit=0, 
                 verbose=1, 
                 reset_base=True,
                 which_base=0):
        """
        Context manager to compute runtime and print result.
        
        string_enter: str
            String printed before timeit starts.
            
        string_exit_callable: Callable[[float], str]
            String printed once timeit ends. It is callable on the time.
            
        level_enter: int (default: 0)
            The enter string is printed if level is lower than verbose.
            
        level_exit: int (default: 0)
            The exit string is printed if level is lower than verbose.
            
        verbose: int (default: 1)
            Threshold to decide if messages are printed. 
            
        reset_base: bool (default: True)
            If True, the time starts with no lag (base = 0). Else the time starts with a lag determine by the previous 
            calls. This is useful if the context manager is inside a loop and the total time to exit the loop must be 
            printed while every iteration of the loop must print a message.
            
        which_base: 0 <= int <= 100
            If several timeit are nested into each other, they should all have a different `which_base`.
        """
        self.string_enter = string_enter
        self.string_exit_callable = string_exit_callable
        self.level_enter = level_enter
        self.level_exit = level_exit
        self.verbose = verbose
        self.reset_base = reset_base
        self.which_base = which_base
        
    def __enter__(self):
        self.start = time.time()
        if self.level_enter <= self.verbose:
            print(self.string_enter, end="")        
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.reset_base:
            TimeIt.base[self.which_base] = 0 
        self.time = (time.time() - self.start) + TimeIt.base[self.which_base]
        TimeIt.base[self.which_base] = self.time
        if self.level_exit <= self.verbose:
            print(self.string_exit_callable(self.time), end="")
        
class RedirectStd:
    
    def __init__(self, where=None):
        """ Redirect the std to the screen, to a file, or discard it.
        
        Argument:
        ---------
        where: str or `sys.stdout` or None (default: None) 
            If str, it is the path to the file where std is written.
            If `sys.stdout`, it is printed on screen.
            If None, std is discarded.
        """
        if where is None: 
            self._out = open(os.devnull, 'w') # do not print
        elif where == sys.stdout:
            self._out = sys.stdout # print on screen
        else:
            self._out = open(where, 'w') # print in file 
            
    def __enter__(self, stdout=None, stderr=None):
        self._old = sys.stdout, sys.stderr
        sys.stdout.flush()
        sys.stderr.flush()
        sys.stdout, sys.stderr = self._out, self._out
        
    def __exit__(self, exc_type, exc_value, traceback):
        self._out.flush()
        sys.stdout, sys.stderr = self._old
        
        
class Formatter:
    
    def __init__(self, width=None, precision=None, fixed=False):
        self.width = width
        self.precision = precision
        self.fixed = fixed
        
    def _fmt(self, v):
        if type(v) == str:
            return v
        elif type(v) == int or type(v) == float or type(v).__name__ == "float64":
            fstr = ''
            if self.width is not None:
                fstr += ' ' + str(self.width)
            if self.precision is not None:
                fstr += '.' + str(self.precision)
            if self.fixed:
                fstr += 'f'
            else:
                fstr += 'g'
            return ('{:' + fstr + '}').format(v)
        elif hasattr(v, '__iter__'):
            return '[' + ','.join(self._fmt(vi) for vi in v) + ']'
        else:
            return str(v)
        
    def __call__(self, *args):
        return '\t'.join(self._fmt(a) for a in args)


# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from typing import List, Callable, Any, Dict, Tuple, Sequence, Union
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt


class Node:

    def __init__(self, *children, **data):
        self._parent = None
        self._children = []
        self._data = data
        self.add(*children)

    @property
    def parent(self):	
        """Parent of the node"""
        return self._parent

    @property
    def children(self):
        """List of children of the node"""
        return self._children
    
    @children.setter
    def children(self, value):
        self._children = value

    @property
    def data(self):
        """Data associated to the node in the form of a dictionary"""
        return self._data

    @data.setter	
    def data(self, value):
        self._data = value

    @data.deleter
    def data(self):
        del self._data
    
    @property
    def is_leaf(self): 
        """True if the node is a tree leaf"""
        return len(self._children) == 0

    @property 
    def is_parent_of_leaf(self):
        if self.is_leaf:
            return False
        else: 
            return self.children[0].is_leaf
        
    @property
    def is_root(self):
         """True if the node is the tree root"""
         return self.parent is None
    
    def make_it_root(self):
        """Turn the node into a root"""
        self._parent = None
        
    @property
    def depth(self):	
        """Number of nodes from the node to the furthest leaf."""
        if self._children:
            return 1 + max(c.depth for c in self._children)
        else:
            return 1

    @property
    def level(self): 
        """Level of the node (root has level 0)"""
        if self._parent:
            return self._parent.level + 1
        else:
            return 0

    @property
    def root(self):
        """Root of the tree (for which self might be a subtree)"""
        return self.branch[0]
    
    @property
    def size(self):	
        """Nomber of nodes in the tree."""
        return 1 + sum(c.size for c in self._children)

    @property
    def address(self):
        if self._parent:
            return self._parent.address + tuple([self._parent._children.index(self)])
        else:
            return tuple([])
    
    @property
    def has_siblings(self):
        return self.parent and len(self.parent.children) >= 2

    @staticmethod
    def get_data_path(node, key, default=None):
        """Returns a dict mapping a level to the data values associated to 'key' on the path leading to a node"""
        return {n.level: n.data.get(key, default) for n in node.branch}

    @property
    def width(self) -> List[int]:
        """Computes the number of nodes on each level (excluding the root)."""
        width = [0] * (self.depth-1)
        for node in self.nodes:
            if not node.is_root:
                width[node.level-1] += 1
        return width
    
    @property
    def bushiness(self) -> List[float]:
        """Returns the list of branching factors at each level 
        (possibly non-integers if structure is non-symmetrical)"""
        return bushiness_from_width(self.width)
    
    @staticmethod
    def bushiness_from_width(width: List[int]) -> List[float]:
        return list(np.array(width) / np.concatenate(([1], width[:-1])))
    
    def has_key(self, 
                key: str, 
                including_levels: List[int] = None, 
                excluding_levels: List[int] = None):
        """ Returns True if the tree rooted at self has a specific data key. 
        
        including_levels: list[int] or None
            Levels to look into. If None, then all levels are looked into except those in excluding_levels.
            
        excluding_levels: list[int] or None
            Levels to be ignored. This argument is used only if `including_levels` is None.
        """
        if excluding_levels is None:
            excluding_levels = []
            
        if including_levels is not None:
            levels_to_look = including_levels
        else:
            levels_to_look = list(range(self.depth))
            for level in excluding_levels:
                levels_to_look.remove(level)
            
        for node in self.nodes:
            if node.level in levels_to_look:
                if key not in node.data.keys():
                    return False
        return True
        
    def node_at_address(self, address):
        """Returns the node located at an address (returns None if no node is at this address)."""
        if address == ():
            return self
        if len(self.children) <= address[0]:
            return None
        else:
            return self.children[address[0]].node_at_address(address[1:])
        
    # --- Iterables ---
    def nodes_at_level(self, t):
        """Returns an iterator on the nodes at level t in the subtree rooted at self"""
        return iter(n for n in self.nodes if n.level == t)

    def ancestor_branches(self, other): #find_split_point??
        """Returns the branches rooted at the closest common ancestor."""
        if self.depth < other.depth:
            return self.find_split_point(other.parent)
        if self.depth > other.depth:
            return self.parent.find_split_point(other)
        if self.parent == other.parent:
            return self, other
        else:
            return self.parent.find_split_branches(self.parent, other.parent)
        
    @property
    def leaves(self):
        """Returns an iterator on the tree leaves."""
        return iter(n for n in self.nodes if n.is_leaf)

    @property
    def branch(self):
        if self._parent:
            return self._parent.branch + [self]
        else:
            return [self]
        
    @property
    def nodes(self):
        """Returns an iterator that walks the tree exhaustively by depth first exploration."""
        yield self
        for c in self._children:
            for w in c.nodes:
                yield w

    @property
    def backward_nodes(self):
        """Returns an iterator that walks the tree exhaustively by backward exploration.
        (i.e., all nodes at level t+1 are explored before any node level at level t)."""
        for t in range(self.depth-1, -1, -1):
            yield from self.nodes_at_level(t)  
        
    @property
    def forward_nodes(self):
        """Returns an iterator that walks the tree exhaustively by breadth first exploration.
        (i.e., all nodes at level t are explored before any node level at level t+1)."""
        for t in range(self.depth):
            yield from self.nodes_at_level(t) 
                
    @property
    def trajectories(self):
        return ((n for n in l.branch) for l in self.leaves)
    
    # --- Structure update ---
    def copy(self, deep_copy=False):
        """Copy the structure of the tree. This copies the references to the nodes and their 
        data dictionary, but the data inside the dictionary still points to the same object 
        (i.e., this is not a deepcopy of the data attribute)."""
        if deep_copy:
            tree = Node(**deepcopy(self.data))
        else:
            tree = Node(**self.data)
        for c in self.children:
            tree.add(c.copy(deep_copy))
        return tree

    def add(self, *children):
        """Adds children to the node."""
        for c in children:
            self._children.append(c)
            c._parent = self
    
    def remove(self):
        """Delete the subtree rooted at self (including self itself)"""
        if self.parent:
            self.parent.children.remove(self)
            self._parent = None
        else:
            raise Exception("cannot remove root")

    def remove_branch(self):
        """Same as remove() but additionally keep deleting nodes backwards along the path 
        leading to self if those nodes don't have children. This deletion ensures that the 
        tree leaves are all at the same level."""
        p = self.parent
        self.remove()
        if p.is_leaf:
            p.remove_branch()
            
    # --- Data update ---        
    def delete_data(self, keys: Union[str, List[str]]):
        """Delete all data under certain keys"""
        if isinstance(keys, str):
            keys = list(keys)
        for node in self.nodes:
            for key in keys:
                if node.data.get(key) is not None:
                    del(node.data[key])
                
    def append_data_dict(self, data_dict):
        """Append data to a tree in place.
        Argument:
        ---------
        data_dict: dict mapping {tuple: dict} 
            Dictionary mapping a node address to a node data (node data is a dictionary)
            Each node in the tree will be appended with the data found at its address in the argument dictionary.
        """    
        for node in self.nodes:
            node.data.update(data_dict.get(node.address, dict()))
          #  for key, value in data_dict.get(node_address, {}).items():
            #    node.data[key] = value
    
    def sort(self, key):
        """Sort (in place) the tree according to some criteria function.
        Argument:
        ---------
        key: function mapping a node to a number.
        """
        self._children.sort(key=key)
        for c in self.children:
            c.sort(key=key)
            
    # --- Encoding ---
    @property            
    def topology(self): 
        """Returns the underlying topology of the tree structure. 
        A topology is the most concise way to encode arbitrary structures."""
        if self.is_leaf:
            pass
        elif self.is_parent_of_leaf: 
            return len(self._children)
        else:
            return tuple([child.topology for child in self._children])

    def get_data_dict(self, with_keys=None, without_keys=None):
        """Turn a tree into a dictionary that maps each node address to its node data.
        
        Argument:
        ---------
        with_keys: list of strings or None
            If not None, it is the data keys to be kept. 
            If None, then all keys in the list will be kept except those in without_keys.
            
        without_keys: list of strings or None
            This argument is used only if 'with_keys' is None. In that case it contains the data keys to be ignored.
        """
        assert with_keys is None or without_keys is None, "One of `with_keys` or `without_keys` must be None"
        if without_keys is None:
            without_keys = []
        if with_keys is not None:
            keys_to_keep = lambda node: set(with_keys)
        else:
            keys_to_keep = lambda node: set(node.data.keys()) - set(without_keys)  
        return {node.address: {key: value for key, value in node.data.items()
                                       if key in keys_to_keep(node)} for node in self.nodes}

    # --- Save and load --- 
    def save_topology(self, path, extension):
        """Save the underlying tree topology into a file.
        This is the best way to save large trees with empty data. 
        For large trees with non-empty data, use .to_file().
        Arguments:
        ----------
        path: string
        
        extension: {'txt', 'pickle'} 
        """
        path = path + f'.{extension}'
        if extension == 'txt':
            with open(path, "w") as f:
                f.write(repr(self.topology))
        elif extension == 'pickle':
            import pickle
            with open(path, "wb") as f:
                pickle.dump(self.topology, f)
        else:
            TypeError(f"Extension should be in {'pickle', 'txt'}, not {extension}.")
            
    def to_file(self, path, extension, with_keys=None, without_keys=None):
        """Save the tree and its data in a file. This is the way to save
        large trees with non-empty data.
        
        Argument:
        ---------
        path: str
        
        extension: {'txt', 'pickle'}
        
        with_keys: list of str or None (default: None)
            List of data keys that will be saved. 
            If None, then all keys in the tree will be saved except those in `without_keys`.
            
        without_keys: list of str or None (default: None)
            List of data keys that will not be saved. 
            This argument is used only if `with_keys` is None.
        """
        if with_keys is not None:
            assert isinstance(with_keys, list), \
            f"Argument `with_keys` should be of type list, not {type(with_keys)}."
        elif without_keys is not None:
            assert isinstance(without_keys, list), \
            f"Argument `without_keys` should be of type list, not {type(without_keys)}."
        data_dict = self.get_data_dict(with_keys, without_keys)
        path = path + f'.{extension}'
        if extension == 'txt':
            np.set_printoptions(threshold=np.inf) # no limit on the number of elements printed in an array
            with open(path, "w") as f:
                f.write(repr(data_dict).replace("},", "},\n"))
        elif extension == 'pickle':
            import pickle
            with open(path, "wb") as f:
                pickle.dump(data_dict, f)
        else:
            TypeError(f"Extension should be 'pickle' or 'txt', not {extension}.")
            
    @classmethod
    def from_file(cls, path, extension):
        """Constructor of a tree and its data directly from a file. 
        This is the way to load large trees with non-empty data. 
        (Reverse of the .to_file() method.)
        Arguments:
        ---------
        path: string
        
        extension: {'txt', 'pickle'}
        """
        path = path + f'.{extension}'
        if extension == 'txt':
            with open(path, "r") as f:
                file_str = f.read()
                file_str = file_str.replace('array', 'np.array')
                file_str = file_str.replace('nan', 'np.nan')
                for type_ in ['int8', 'int16', 'int64', 'float16', 'float32']:
                    file_str = file_str.replace(f'dtype={type_}', f'dtype="{type_}"')
                data_dict = eval(file_str)
        elif extension == 'pickle':
            import pickle
            with open(path, "rb") as f:
                data_dict = pickle.load(f)
        return cls.from_data_dict(data_dict)
    
    # --- Alternative constructors ---
    @classmethod
    def from_data_dict(cls, data_dict: Dict[Tuple[int], dict]):
        """Constructor of a tree from a dictionary that maps each node address to its node data.
        It is the reverse of the .get_data_dict() method, i.e., tree = from_data_dict(tree.get_data_dict())"""
        tree = cls._from_addresses(list(data_dict.keys()))    
        for node in tree.nodes:
            node.data = data_dict.get(node.address, {})
        return tree
    
    @classmethod
    def _from_addresses(cls, addresses):
        """
        Constructor of a tree structure from a list of node addresses.

        Arguments:
        ----------
        addresses: list of tuples
            Each tuple corresponds to the address of a node in the structure (as would be given 
            by node.address()).
        
        Returns:
        --------
        Node: the root node of the specified tree structure
        """
        root = cls()
        for address in addresses:
            if root.node_at_address(address) is not None:
                continue
            else:
                root.node_at_address(address[:-1]).add(cls())  
        return root

    @classmethod
    def from_topology(cls, topology: Sequence[Sequence[int]], _tree=None):
        """
        Constructor of a tree structure from a topology. 
        A topology is the most general way to build any possible type of structure.
        
        Arguments:
        ----------
        topology: tuple/list of int nested into (possibly several) layers of tuple/list.
            The integers are the number of sibling leaves. Integers (and sub-tuples) are inside 
            the same tuple if they have the same parent. 
        
        _tree: unused argument (internally used for the recurrent calls to the method)
            
        Returns:
        --------
        Node: the root node of the specified tree structure
        """
        if _tree is None:
            _tree = cls()
        if isinstance(topology, int):
            _tree.add(*[cls() for i in range(topology)])
            return _tree
        else: 
            if _tree.is_leaf:
                _tree.add(*[cls() for i in range(len(topology))]) 
            for item, leaf in zip(topology, list(_tree.leaves)):
                cls.from_topology(item, leaf)
            return _tree
        
    @classmethod
    def from_recurrence(cls, 
                        max_level: int, 
                        init: int, 
                        recurrence: Dict[int, Sequence[int]], 
                        _tree=None):
        """
        Constructor of a tree structure where the number of child nodes satisfies a recurrence relation.
        
        Arguments:
        ----------
        max_level: integer
            Level of the tree leaves (root has level 0)
            
        init: integer
            Number of nodes at level 1 (to initialize the recurrence).
            
        recurrence: dictionary that maps an integer k to a k-tuple of integers 
            All integers must be positive and correspond to the number of sibling nodes. 
            (Each integers found in the dict values must also be found in the keys.)
            
        _tree: unused argument (internally used for the recurrent calls to the method)

        Example:     
            recurrence = {1: (2,), 2: (1,3), 3: (1,2,3)}) 
            will produce a tree where:
                - a single node (without sibling) will give birth to 2 child nodes,
                - 2 sibling node will give birth to 1 and 3 children, 
                - 3 sibling nodes will give birth to 1, 2, and 3 children.

        Returns:
        --------
        Node: the root node of the specified tree structure
        """
        if _tree is None:
            _tree = cls(*[cls() for _ in range(init)])
        else:
            _tree.add(*[cls() for _ in range(init)])
        if max_level == 1:
             return _tree
        else:
            for i, node in enumerate(_tree.children):
                if node.is_leaf:
                    cls.from_recurrence(max_level-1, recurrence[len(_tree.children)][i], recurrence, node)
            return _tree
    
    @classmethod
    def from_bushiness(cls, bushiness: Sequence[int]):
        """
        Constructor of a tree structure with a time-dependent branching factor.
        
        Arguments:
        ----------
        bushiness: tuple of integers
            Integer at location i in the tuple is the branching factor at the i-th level
            
        Returns:
        --------
        Node: the root node of the specified tree structure
        """
        integer_bushiness = np.array(bushiness).astype('int')
        assert (np.array(bushiness) == integer_bushiness).all(), \
            f"The bushiness must only contain integer values, not {bushiness}"
        tree = cls()
        for t in range(len(integer_bushiness)):
            for leaf in list(tree.leaves):
                leaf.add(*[cls() for k in range(integer_bushiness[t])])
        return tree
    
    # --- Plots ---
    def plot(self, 
             print_on_nodes: Callable[['Node'], str] = None,
             print_on_edges: Callable[['Node'], str] = None,
             max_stage=None,
             centered_around_leaves=True, 
             sorting_fct=None,
             color_nodes='blue',
             size_nodes=20,
             color_edges='blue',
             width_edges=0.5,
             size_txt_nodes=12,
             color_txt_nodes='black',
             size_txt_edges=10,
             color_txt_edges='black',
             figsize=(5,5),
             xlim=None,
             ylim=None,
             show_xaxis=False,
             show_yaxis=False,
             show_frame=False,
             to_file=None,
             extension=".pdf",
             ax=None):
    
        if sorting_fct is not None:
            tree = self.copy()
            tree.sort(sorting_fct)
        else: 
            tree = self
            
        if max_stage is not None:
            tree = tree.copy()
            for node in list(tree.nodes_at_level(max_stage+2)):
                node.remove()
        else:
            max_stage = tree.depth
            
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        
        # define the color and size functions for each node and edge
        color_nodes_fct = lambda node: color_nodes if not callable(color_nodes) else color_nodes
        size_nodes_fct = lambda node: size_nodes if not callable(size_nodes) else size_nodes
        color_edges_fct = lambda node: color_edges if not callable(color_edges) else color_edges 
        width_edges_fct = lambda node: width_edges if not callable(width_edges) else width_edges 
        size_txt_nodes_fct = lambda node: size_txt_nodes if not callable(size_txt_nodes) else size_txt_nodes
        color_txt_nodes_fct = lambda node: color_txt_nodes if not callable(color_txt_nodes) else color_txt_nodes
        size_txt_edges_fct = lambda node: size_txt_edges if not callable(size_txt_edges) else size_txt_edges
        color_txt_edges_fct = lambda node: color_txt_edges if not callable(color_txt_edges) else color_txt_edges
        
        if print_on_nodes is None:
            print_on_nodes = lambda node: None
        if print_on_edges is None:
            print_on_edges = lambda node: None
            
        # get the vertical position for the leaves
        pos_y = {leaf.address: i for i, leaf in enumerate(tree.leaves)}
    
        # get the vertical position for all the nodes but the leaves
        for node in tree.backward_nodes:
            if not node.is_leaf:
                if centered_around_leaves:
                    pos_y[node.address] = np.mean([pos_y[n.address] for n in node.leaves])
                else:
                    pos_y[node.address] = np.mean([pos_y[n.address] for n in node.children])
                    
        # plot the nodes
        for node in tree.nodes:
            if node.level == max_stage+1:
                continue
            x, y = node.level, pos_y[node.address]
            ax.scatter(x, y, color=color_nodes_fct(node), s=size_nodes_fct(node))
            
            # write text on node if required
            txt = print_on_nodes(node)
            if txt is not None:
                fontsize = size_txt_nodes_fct(node)
                color = color_txt_nodes_fct(node)
                if not node.parent:
                    ax.text(x-0.1, y, txt, fontsize=fontsize, color=color, 
                            verticalalignment='center', horizontalalignment='right')
                elif node.is_leaf:
                    ax.text(x+0.1, y, txt, fontsize=fontsize, color=color,  
                            verticalalignment='center', horizontalalignment='left')
                else:
                    ax.text(x-0.05, y, txt, fontsize=fontsize, color=color, 
                            verticalalignment='center', horizontalalignment='right')
    
        # plot the edges
        for node in tree.nodes:
            if node.level == max_stage+1:
                continue
                
            if node.parent:
                x = [node.level, node.parent.level]
                y = [pos_y[node.address], pos_y[node.parent.address]]
                ax.plot(x, y, color=color_edges_fct(node), linewidth=width_edges_fct(node))
                
                # write text on edge if required
                txt = print_on_edges(node)
                if txt is not None:
                    ax.text(np.mean(x), np.mean(y), txt, 
                                fontsize=size_txt_edges_fct(node), color=color_txt_edges_fct(node), 
                                verticalalignment='center', horizontalalignment='center')
            
        if xlim is not None:
            ax.set_xlim(*xlim)
        if ylim is not None:
            ax.set_ylim(*ylim)
        ax.get_xaxis().set_visible(show_xaxis)
        ax.get_yaxis().set_visible(show_yaxis)
        if not show_frame and not show_xaxis and not show_yaxis:
            ax.axis('off')
            
        if to_file is not None:
            plt.savefig(to_file + extension)
        return ax

    @staticmethod
    def plot_multiple_trees(*trees, along_x=True, figsize=(5,5), to_file=None, extension=".pdf", **kwargs):
        """
        Plots several trees side by side.
        
        Arguments:
        ----------
        *trees: instances of Node
            All trees to be plotted.
            
        along_x: boolean
            If True, trees are displayed from left to right; if False, from up to down.
            
        figsize: 2-tuple of numbers
            Size of the whole figure (not of each figure individually).
            
        to_file: string
            Path where the figure will be saved (not saved if None).
           
        extension: string
            Extension in which the figure will be saved.
            
        **kwargs: 
            All keyword args of .plot() except 'ax', 'figsize' and 'to_file'.
        """
        assert len(trees) > 1, "There must be 2 trees at least"
        kwargs.pop('ax', None), kwargs.pop('figsize', None), kwargs.pop('to_file', None)
        if along_x:
            fig, axes = plt.subplots(1, len(trees), figsize=figsize)
        else:
            fig, axes = plt.subplots(len(trees), 1, figsize=figsize)
            
        for index, tree in enumerate(trees):
            axes[index] = tree.plot(ax=axes[index], **kwargs)
            
        if to_file is not None:
            plt.savefig(to_file + extension)
        plt.show()
        
    # --- Representations ---
    def __repr__(self):
        return self.format()

    def format(self, level=0, margin=20, label=lambda n: "Node", data=None):
       # from stochoptim.util import Formatter
        if data is None:
            ndata = self.data.keys()
        else:
            ndata = data
        #fmt = Formatter(precision=3)
        s1 = '\n' + ('  ' * level) + str(label(self))
        s1 += ' ' * (margin - len(s1))
        #s1 += '\t'.join('{}={}'.format(k, fmt(self.data[k])) for k in ndata if k in self.data)
        s1 += '\t'.join('{}={}'.format(k, self.data[k]) for k in ndata if k in self.data)
        s1 += ''.join(c.format(level + 1, margin, label, data) for c in self.children)
        if level == 0:
            s1 = s1[1:]
        return s1

get_data_path = Node.get_data_path
bushiness_from_width = Node.bushiness_from_width
from_file = Node.from_file
from_data_dict = Node.from_data_dict
from_topology = Node.from_topology
from_recurrence = Node.from_recurrence
from_bushiness = Node.from_bushiness